"""
Code generation for CVXMOD.

"""

# Copyright (C) 2006-2008 Jacob Mattingley and Stephen Boyd.
#
# This file is part of CVXMOD.
#
# CVXMOD is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# CVXMOD is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from cvxmod.base import *
from cvxmod.base import withbrackets, addfunction, multfunction, Jwarn,\
        addormultfunction, symbslice
from cvxmod.symbolic import *
from cvxmod.symbolic import concatvertsymb, concathorizsymb, constsymb, eyesymbol
from cvxmod.util import bylowerstr, joinlist
from cvxopt.amd import order

from time import strftime
import os
import re

class nze(symbol):
    # Nonzero elements, eg for use in a matrix. Invariant under multiplication
    # and division, except for by zero.
    def __init__(self, name=None):
        self.rows = 1
        self.cols = 1
        self.name = name

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.name is None:
            return '<nze>'
        else:
            return '<nze %s>' % self.name

    def __div__(self, other):
        if iszero(other):
            raise ZeroDivisionError
        else:
            return self

    def __rdiv__(self, other):
        if iszero(other):
            return 0
        else:
            return self

    def __mul__(self, other):
        if iszero(other):
            return 0
        else:
            return self

    def __rmul__(self, other):
        if iszero(other):
            return 0
        else:
            return self

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

class divfunction(addormultfunction):
    # first cut. no size checking or anything.
    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs
        self.rows = 1
        self.cols = 1
        self.brackets = True

    def __str__(self):
        return withbrackets(self.lhs) + '/' + withbrackets(self.rhs)

def divide(lhs, rhs):
    if iszero(lhs) or iszero(rhs):
        return 0
    else:
        return divfunction(lhs, rhs)

class sparsedict(dict):
    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            return 0

def addsplit(obj):
    # Argument order is (pre-operations, actual, nonindexed).
    if isoptvar(obj) or isparam(obj) or isinstance(obj, (int, float)):
        if is1x1(obj):
            return ([], [], obj)
        else:
            return ([], [obj], 0)
    elif isinstance(obj, addfunction):
        return [x + y for (x, y) in zip(addsplit(obj.lhs), addsplit(obj.rhs))]
    elif isinstance(obj, multfunction):
        if is1x1(obj.lhs):
            (pre, actual, nonind) = addsplit(obj.rhs)
            pre = [obj.lhs*x for x in pre]
            actual = [obj.lhs*x for x in actual]
            nonind = obj.lhs*nonind
            
            return (pre, actual, nonind)
        else:
            raise NotImplementedError('could not handle ' + str(obj))
    elif isinstance(obj, constsymb):
        return ([], [], obj.constval)
    else:
        raise NotImplementedError('could not handle ' + str(obj))

def stuffmatrix(obj, name):
    s = ''
    s += '\n// Begin: definition of %s.\n' % name
    for l in strsymbmatrix(obj, name).splitlines():
        s += '//   %s\n' % l

    s += '%s = malloc((%s)*sizeof(double));\n\n' % \
            (name, str(rows(obj) * cols(obj)))

    s += stuffentry(obj, name, rows(obj), 0, 0, 0)
    s += '// End: definition of %s.\n' % name

    return s

def stuffentry(obj, name, rs, csleft, rsabove, rsbelow, smult=1):
    def stuff1x1(val):
        return '%s[%s] = %s;\n' % (name, str(topleft), str(val))

    def nwsm(n):
        # Name with scalar multiplier.
        # Takes smult implicitly.
        return '%s' % str(smult*n)

    # here (or elsewhere) keep track of existing multiplied stuff. for example,
    # if we have tr(c)*x, don't want to have to work it out every time.

    # Everything should be able to accept a scalar multiple for one of its
    # arguments.

    s = ''
    topleft = rs*csleft + rsabove

    if isinstance(obj, (int, float, matrix, spmatrix)) and is1x1(obj):
        s += stuff1x1(str(obj))
    elif isinstance(obj, concatvertsymb):
        s += stuffentry(obj.lhs, name, rs, csleft, rsabove, rsbelow, smult)
        s += stuffentry(obj.rhs, name, rs, csleft, rsabove + rows(obj.lhs),
                        rsbelow - rows(obj.lhs), smult)
    elif isinstance(obj, concathorizsymb):
        s += stuffentry(obj.lhs, name, rs, csleft, rsabove, rsbelow, smult)
        s += stuffentry(obj.rhs, name, rs, csleft + cols(obj.lhs), rsabove,
                        rsbelow, smult)
    elif isinstance(obj, constsymb):
        s += 'for (col = 0; col < %s; col++)\n' % str(cols(obj))
        s += ' '*4 + 'for (row = 0; row < %s; row++)\n' % str(rows(obj))
        s += ' '*8 + '%s[%s + col*(%s) + row] = %s;\n\n' % \
                (str(name), str(topleft), str(rs), nwsm(obj.constval))
    elif isinstance(obj, eyesymbol):
        s += 'for (col = 0; col < %s; col++)\n' % str(cols(obj))
        s += ' '*4 + 'for (row = 0; row < %s; row++)\n' % str(rows(obj))
        s += ' '*8 + '%s[%s + col*(%s) + row] =\n          (row == col) ? %s : 0;\n\n' % \
                (str(name), str(topleft), str(rs), str(smult))
    elif isparam(obj):
        s += 'for (col = 0; col < %s; col++)\n' % str(cols(obj))
        s += ' '*4 + 'for (row = 0; row < %s; row++)\n' % str(rows(obj))
        s += ' '*8 + '%s[%s + col*(%s) + row] =\n          %s[col*(%s) + row];\n\n' % \
                (str(name), str(topleft), str(rs), nwsm(obj), str(rows(obj)))
    elif isinstance(obj, multfunction):
        # nastiness of non-recursion here!
        if is1x1(obj.lhs):
            s += stuffentry(obj.rhs, name, rs, csleft, rsabove, rsbelow, obj.lhs)
        else:
            raise NotImplementedError('could not handle ' + str(obj))
    elif isinstance(obj, addfunction):
        # pre: temporary variables that need to be calculated first.
        # (need to come up with an indexing scheme so that A*x, for example,
        # can be reused. Probably assume 
        # actual: pieces of expression that should be indexed.
        # noniter: pieces of expression that should not be indexed.

        (pre, actual, nonind) = addsplit(obj)

        if pre:
            s += pre + '\n\n'

        rhs = "\n" + " "*11 + " + " .join(
            [str(smult*x) + ("[col*(%s) + row]" % str(cols(x))) for x in actual])
        if nonind:
            rhs += "\n" + " "*10 + " + " + str(smult*nonind)

        s += 'for (col = 0; col < %s; col++)\n' % str(cols(obj))
        s += ' '*4 + 'for (row = 0; row < %s; row++)\n' % str(rows(obj))
        s += ' '*8 + '%s[%s + col*(%s) + row] = %s;\n' % \
                (str(name), str(topleft), str(rs), rhs)
    else:
        raise NotImplementedError('could not handle ' + str(obj))

    return s

def exprtoC(obj):
    if isinstance(obj, symbslice):
        (a, b) = obj.splitsl()
        if b is None:
            return str(obj.arg) + '[%d]' % a
        else:
            index = a + b*value(rows(obj.arg))
            return str(obj.arg) + '[%d]' % index
    elif isinstance(obj, (int, float)):
        return str(obj)
    elif isinstance(obj, multfunction):
        if obj.lhs is -1:
            return '-' + exprtoC(obj.rhs)
        else:
            return exprtoC(obj.lhs) + '*' + exprtoC(obj.rhs)
    elif isinstance(obj, addfunction):
        return exprtoC(obj.lhs) + ' + ' + exprtoC(obj.rhs)
    else:
        raise NotImplementedError('cannot handle %s' % str(obj))

class codegen(object):
    def __init__(self, p, outdir, skel):
        self.p = p
        self.outdir = outdir
        self.skel = skel

    def printmatrix(self, x):
        m = rows(x)
        n = cols(x)
        s = 'printf("%s =\\n");\n' % str(x)
        s += "printmatrix(%s, %s, %s);\n" % (str(x), str(m), str(n))
        self.write(s)

    def declarematrix(self, A):
        s = "double %s[] = {\n " % str(A)
        offset = 1
        v = matrix(value(A))
        k = 0
        for j in range(value(cols(A))):
            for i in range(value(rows(A))):
                t = "% .7g, " % v[i,j]
                # jem: enhance this later to be nicer looking.
                s += t.rjust(12)
                if (k % 7) == 6:
                    s += "\n"
                    s += ' ' * offset
                k += 1
            #s += "\n"
            #s += ' ' * offset

        s = s[:-(offset+3)] + "};\n\n"
        return s

    def readme(self):
        p = self.p

        d = {}
        d['header'] = header()
        d['probdef'] = str(p)
        d['stdform'] = p.strsymbsolve()

        writecode(self.outdir, self.skel, 'README', d)

    def tostdform(self):
        p = self.p

        ((ct, At, bt, Gt, ht), xt, d, optvars) = p.symbsolve('cAbGh')

        d = {}
        d['header'] = header()

        # Now is when we determine the signature.
        sig = 'lpstdform tostdform(\n' + ' '*8
        if getdims(p):
            sig += ', '.join(['int %s' % x for x in bylowerstr(getdims(p))])

        if getparams(p):
            sig += ',\n' + ' '*8
            sig += ', '.join(['double *%s' % x for x in bylowerstr(getparams(p))])

        sig += ')'
        d['tostdform_sig'] = sig

        d['lsf_c'] = stuffmatrix(ct, 'lsf->c')
        d['lsf_A'] = stuffmatrix(At, 'lsf->A')
        d['lsf_b'] = stuffmatrix(bt, 'lsf->b')
        d['lsf_G'] = stuffmatrix(Gt, 'lsf->G')
        d['lsf_h'] = stuffmatrix(ht, 'lsf->h')

        sig = 'int convertback(outvar *vars,\n' + ' '*8
        if getdims(p):
            sig += ', '.join(['int %s' % x for x in bylowerstr(getdims(p))])
        sig += ',\n        double *xt)'

        d['convertback_sig'] = sig
        d['lsf_m'] = str(rows(At))
        d['lsf_n'] = str(cols(At))
        d['lsf_p'] = str(rows(Gt))

        s = ''
        i = 0
        rs = 0
        for x in optvars:
            if x in getoptvars(p):
                # jem: need to also deal with substitute back situations.
                s += '// Recover %s.\n' % str(x)
                name = 'vars[%d]' % i;
                s += '%s->m = %s;\n' % (name, str(rows(x)))
                s += '%s->n = 1;\n' % name # for the moment (jem).
                s += '%s->val = malloc((%s)*sizeof(double));\n' % (name, str(cols(x)))
                s += 'for (row = 0; row < %s; row++)\n' % str(rows(x))
                s += '    %s->val[row] = xt[%s + row];\n\n' % (name, str(rs))
                i += 1

            rs = compactdims(rs + rows(x))

        d['varrecover'] = s

        writecode(self.outdir, self.skel, 'tostdform.h', d)
        writecode(self.outdir, self.skel, 'tostdform.c', d)

    def data(self):
        p = self.p

        d = {}
        d['header'] = header()

        s = ''
        for x in getparams(p):
            s += self.declarematrix(x)
        d['params'] = s

        writecode(self.outdir, self.skel, 'data.h', d)

    def feasdata(self, As, bs, n, p, LOWER):
        d = {}
        d['header'] = header()


        print 'Solving feasibility problem to get feasible starting point.'
        As = value(As)
        bs = value(bs)
        p = value(p)
        x = optvar('x', n)
        prob = problem(constr=[As*x == bs, x[:p] >= 0, x[p:] >= -LOWER + 1])
        prob.solve()

        x = value(x)
        s = ''
        for i in range(rows(x)):
            s += 'CM_x[%d] = % .8g;\n' % (i, x[i])

        d['CM_x'] = s
        writecode(self.outdir, self.skel, 'feas.h', d)

        return x

    def testfa(self, n):
        p = self.p

        d = {}
        d['header'] = header()

        d['n'] = n

        d['params'] = joinlist(bylowerstr(getparams(p)), True)
        d['optvars'] = joinlist(bylowerstr(getoptvars(p)), True)

        s = joinlist(bylowerstr(getparams(p)), True)
        d['fa_args'] = s

        writecode(self.outdir, self.skel, 'testfa.c', d)

    def testba(self, n):
        p = self.p

        dt = {}
        dt['header'] = header()

        dt['n'] = n

        dt['params'] = joinlist(bylowerstr(getparams(p)), True)
        dt['optvars'] = joinlist(bylowerstr(getoptvars(p)), True)

        print 'Creating method signature.'
        if getparams(self.p):
            dt['fb_sig_params'] = ', '.join(['double *%s' % x for x in \
                                     bylowerstr(getparams(self.p))])
            dt['fb_sig_params'] += ','
        else:
            dt['fb_sig_params'] = ''

        dt['fb_args'] = joinlist(bylowerstr(getparams(p)), True)

        writecode(self.outdir, self.skel, 'testba.c', dt)

    def test2p(self, As, bs, cs, m, n, p):
        dt = {}
        dt['header'] = header()
        dt['m'] = value(m)
        dt['n'] = value(n)
        dt['p'] = value(p)

        dt['params'] = joinlist(bylowerstr(getparams(self.p)), True)
        dt['optvars'] = joinlist(bylowerstr(getoptvars(self.p)), True)

        print 'Generating code for two phase fast barrier method:'

        print 'Creating method signature.'
        if getparams(self.p):
            dt['fb_sig_params'] = ', '.join(['double *%s' % x for x in \
                                     bylowerstr(getparams(self.p))])
            dt['fb_sig_params'] += ','
        else:
            dt['fb_sig_params'] = ''

        A = nzentries(As)
        b = nzentries(bs)
        print 'Generating code for initial feasible point.'
        s = ''
        # CVXMOD_WL_t1.
        t1 = optvar('CM_t1', m, 1)
        for i in range(m):
            t = -b[i]
            for j in range(n):
                if (i,j) in A:
                    t = t + A[i,j]
            s += 'CM_t1[%d] = %s;\n' % (i, exprtoC(t))
        dt['CM_t1'] = s
        dt['fb_args'] = joinlist(bylowerstr(getparams(self.p)), True)

        writecode(self.outdir, self.skel, 'test2p.c', dt)


    def fastacent(self, As, cs, m, n, p):
        dt = {}
        dt['header'] = header()
        dt['m'] = value(m)
        dt['n'] = value(n)
        dt['p'] = value(p)

        print 'Generating code for fast analytic centering method:'

        print 'Framing objective.'
        cs = value(cs)
        s = ''
        for i in range(n):
            s += 'CM_c[%d] = %s;\n' % (i, exprtoC(cs[i]))
        dt['CM_c'] = s

        print 'Creating method signature.'
        if getparams(self.p):
            dt['params'] = ', '.join(['double *%s' % x for x in \
                                      bylowerstr(getparams(self.p))])
            dt['params'] += ','
        else:
            dt['params'] = ''

        print 'Inspecting problem structure.'
        A = nzentries(As)
        AT = nzentries(tp(As))

        # Create a gh optvar simply to make it easy to convert the multiplication.
        gh = optvar('CM_gh', n, 1)
        
        # Now multiply and solve.
        print 'Generating reduced Newton system coefficient matrix.'
        (ASATnum, s, ss, p) = self.multstuff(A, AT, m, n)
        dt['CM_ASAT'] = s
        dt['CM_ASAT_size'] = len(ss)

        # Reverse the permutation. Ran into a weird bug when trying a fancier
        # scheme of inverting the permutation.
        pinv = [None,]*len(p)
        for i in range(len(p)):
            pinv[p[i]] = i

        print 'Generating code for b = A*gh.'
        s = ''
        # Take note of the (inverse) permutation p.
        for i in range(m):
            d = 0
            for j in range(n):
                if (p[i], j) in A:
                    d = d - A[p[i],j]*gh[j]
            s += 'CM_b[%d] = %s;\n' % (i, exprtoC(d))
        dt['CM_b'] = s


        print 'Generating Cholesky factorization and solution code:'
        (d, (L, ssL, KL)) = self.cholsolve(ASATnum, ss)

        dt.update(d)

        print 'Generating code for dx.'
        s = ''
        w = optvar('CM_w', m, 1)
        for i in range(n):
            d = 0
            for j in range(m):
                if (i, j) in AT:
                    # Take note of the (inverse) permutation p.
                    d = d + AT[i,j]*w[pinv[j]]
            s += 'CM_dx[%d] = -CM_h[%d]*(%s + CM_g[%d]);\n' % (i, i, exprtoC(d), i)
        dt['CM_dx'] = s

        writecode(self.outdir, self.skel, 'fastacent.c', dt)

    def fastbarr(self, As, bs, cs, m, n, p, suffix=''):
        dt = {}
        dt['header'] = header()
        dt['m'] = value(m)
        dt['n'] = value(n)
        dt['p'] = value(p)

        dt['suffix'] = suffix

        print 'Generating code for fast barrier method:'

        print 'Framing objective.'
        cs = value(cs)
        s = ''
        for i in range(n):
            s += 'CM_c[%d] = %s;\n' % (i, exprtoC(cs[i]))
        dt['CM_c'] = s

        print 'Creating method signature.'
        if getparams(self.p):
            dt['params'] = ', '.join(['double *%s' % x for x in \
                                      bylowerstr(getparams(self.p))])
            dt['params'] += ','
        else:
            dt['params'] = ''

        print 'Inspecting problem structure.'
        A = nzentries(As)
        AT = nzentries(tp(As))

        # Create a gh optvar simply to make it easy to convert the multiplication.
        gh = optvar('CM_gh', n, 1)
        
        # Now multiply and solve.
        print 'Generating reduced Newton system coefficient matrix.'
        (ASATnum, s, ss, p) = self.multstuff(A, AT, m, n)
        dt['CM_ASAT'] = s
        dt['CM_ASAT_size'] = len(ss)

        # Reverse the permutation. Ran into a weird bug when trying a fancier
        # scheme of inverting the permutation.
        pinv = [None,]*len(p)
        for i in range(len(p)):
            pinv[p[i]] = i

        print 'Generating code for b = A*gh.'
        s = ''
        # Take note of the (inverse) permutation p.
        for i in range(m):
            d = 0
            for j in range(n):
                if (p[i], j) in A:
                    d = d - A[p[i],j]*gh[j]
            s += 'CM_b[%d] = %s;\n' % (i, exprtoC(d))
        dt['CM_b'] = s


        print 'Generating Cholesky factorization and solution code:'
        (d, (L, ssL, KL)) = self.cholsolve(ASATnum, ss)

        dt.update(d)

        print 'Generating code for dx.'
        s = ''
        w = optvar('CM_w', m, 1)
        for i in range(n):
            d = 0
            for j in range(m):
                if (i, j) in AT:
                    # Take note of the (inverse) permutation p.
                    d = d + AT[i,j]*w[pinv[j]]
            s += 'CM_dx[%d] = -CM_h[%d]*(%s + CM_g[%d]);\n' % (i, i, exprtoC(d), i)
        dt['CM_dx'] = s

        writecode(self.outdir, self.skel, 'fastbarr.c', dt,
                  outname='fastbarr%s.c' % suffix)

    def testwithdata(self, dispvars=False):
        p = self.p

        d = {}
        d['header'] = header()

        d['dims'] = joinlist(bylowerstr(getdims(p)), True)
        d['params'] = joinlist(bylowerstr(getparams(p)), True)
        d['optvars'] = joinlist(bylowerstr(getoptvars(p)), True)
        d['numvars'] = str(len(getoptvars(p)))

        # jem: making the assumption that optvars order will be sorted like
        # this.
        s = ''
        i = 0
        for x in bylowerstr(getoptvars(p)):
            s += 'outvar %s;\n%s = vars[%d];\n' % (str(x), str(x), i)
            i += 1
        d['namevars'] = s

        s = joinlist(bylowerstr(getdims(p)) + bylowerstr(getparams(p)), True)
        d['tsf_args'] = s

        # jemjemjem.
        #if dispvars:
        #    d['tsf_args'] = s
        #// Display the variables.
        #pm(x->val, x->m, x->n);
        #pm(y->val, y->m, y->n);

        writecode(self.outdir, self.skel, 'testwithdata.c', d)

    def backsubs(self, L, n, ssL):
        # Include an implicit transpose here. ie, L is a lower triangular
        # version of an upper triangular matrix.
        s = ''
        for i in range(n - 1, -1, -1):
            #s += 'CM_w[%d] = CM_x[%d];\n' % (i, i)
            for j in range(i + 1, n):
                if not iszero(L[j,i]):
                    s += 'CM_w[%d] -= CM_L[%s]*CM_w[%d];\n' % (i, ssL[j,i], j)

        return s

    def forwardsubs(self, L, n, ssL):
        s = ''
        for i in range(n):
            s += 'CM_w[%d] = CM_b[%d];\n' % (i, i)
            for j in range(0, i):
                if not iszero(L[i,j]):
                    s += 'CM_w[%d] -= CM_L[%s]*CM_w[%d];\n' % (i, ssL[i,j], j)

        return s

    def cholsolve(self, Anum, ssA):
        n = rows(Anum)
        (L, ssL, KL, Aclean) = self.cholss(Anum, ssA)

        dt = {}
        dt['CM_L_size'] = KL

        print '- code for Cholesky factorization.'
        dt['chol'] = self.chol(Aclean, n, ssA, ssL)
        print '- code for forward substitution.'
        dt['forwardsubs'] = self.forwardsubs(L, n, ssL)

        print '- code for back substitution.'
        dt['backsubs'] = self.backsubs(L, n, ssL)
        return (dt, (L, ssL, ssA))

    def chol(self, A, n, ssA, ssL):
        v = [None,]*n
        d = [None,]*n
        L = {}
        d[0] = nze()
        s = 'CM_d[0] = CM_ASAT[%d];\n' % ssA[0,0]

        for i in range(1, n):
            L[i,0] = A[i,0]# / d[0]
            if (i,0) in ssA:
                s += 'CM_L[%d] = CM_ASAT[%d] / CM_d[0];\n' % (ssL[i,0], ssA[i,0])

        for j in range(1, n):
            for i in range(j):
                v[i] = L[j,i]*d[i]
                if isnonzero(L[j,i]) and isnonzero(d[i]):
                    s += 'CM_v[%d] = CM_L[%d]*CM_d[%d];\n' % (i, ssL[j,i], i)

            v[j] = A[j,j]
            s += 'CM_v[%d] = CM_ASAT[%d];\n' % (j, ssA[j,j])

            for i in range(j):
                v[j] = v[j] - L[j,i]*v[i]
                if isnonzero(L[j,i]*v[i]):
                    s += 'CM_v[%d] -= CM_L[%d]*CM_v[%d];\n' % (j, ssL[j,i], i)

            d[j] = v[j]
            s += 'CM_d[%d] = CM_v[%d];\n' % (j, j)

            if j < n - 1:
                for i in range(j + 1, n):
                    L[i,j] = A[i,j]
                    if (i,j) in ssL:
                        if isnonzero(A[i,j]):
                            s += 'CM_L[%d] = CM_ASAT[%d];\n' % (ssL[i,j], ssA[i,j])
                        else:
                            s += 'CM_L[%d] = 0;\n' % ssL[i,j]
                    else:
                        continue

                    for k in range(j):
                        L[i,j] = L[i,j] - L[i,k]*v[k]
                        if isnonzero(L[i,k]*v[k]):
                            s += 'CM_L[%d] -= CM_L[%d]*CM_v[%d];\n' % (ssL[i,j], ssL[i,k], k)

                    if isnonzero(L[i,j]):
                        s += 'CM_L[%d] /= CM_v[%d];\n' % (ssL[i,j], j)

        # Implicitly assuming CM_L_{ii} = 1 for all i.
        return s

    def cholss(self, Anum, ssA):
        # Returns (L, Lss) where Lss[i,j] returns the index to store L[i,j].
        # L[i,j] = nze() or 0.
        # Note: implicit about the fact that L_{ii} = 1 for all i.
        n = rows(Anum)

        # Work from a clean A.
        A = sparsedict()
        for i in range(n):
            for j in range(i + 1):
                if isnonzero(Anum[i,j]):
                    A[i,j] = nze()

        v = [None,]*n
        d = [None,]*n
        L = sparsedict()
        d[0] = A[0,0]

        for i in range(1, n):
            L[i,0] = A[i,0]

        for j in range(1, n):
            for i in range(j):
                v[i] = L[j,i]*d[i]

            v[j] = A[j,j]

            for i in range(j):
                v[j] = v[j] - L[j,i]*v[i]

            d[j] = v[j] # will probably be translated directly.

            if j < n - 1:
                for i in range(j + 1, n):
                    L[i,j] = A[i,j]
                    for k in range(j):
                        L[i,j] = L[i,j] - L[i,k]*v[k]

        # Compute (simple) storage scheme for L.
        k = 0
        ssL = {}
        Lclean = sparsedict()
        for i in range(n):
            for j in range(n):
                if isnonzero(L[i,j]):
                    ssL[i,j] = k
                    Lclean[i,j] = nze()
                    k += 1
        KL = k

        return (Lclean, ssL, KL, A)

    def cholmoddata(self, Anum):
        # jem: temporary function.
        # jem: later need to store only one half of A: pick which!

        d = {}
        d['header'] = header()

        # There must be a better way of doing this.
        n = rows(Anum)
        s = ''
        s += 'int *Ai = At->i;\n'
        s += 'int *Aj = At->j;\n'
        s += 'double *Ax = At->x;\n'

        k = 0
        for i in range(n):
            for j in range(i, n):
                if isnonzero(Anum[i,j]):

                    s += 'Ai[%d] = %d; ' % (k, i)
                    s += 'Aj[%d] = %d; ' % (k, j)
                    s += 'Ax[%d] = %.7g;\n' % (k, Anum[i,j])
                    k += 1

        s = 'At = cholmod_allocate_triplet(%d, %d, %d, 1, CHOLMOD_REAL, &common);\n' % \
                (n, n, k) + s
        s = 'cholmod_triplet *At;\n' + s
        s += 'At->nnz = %d;\n' % k

        d['At'] = s

        writecode(self.outdir, self.skel, 'cholmod_data.h', d)

    def choldata(self, Anum, bnum, n):
        # jem: later need to store only one half of A: pick which!

        d = {}
        d['header'] = header()

        ssA = self.cholss(Anum, True)

        s = ''
        for (i, j) in sorted(ssA.keys()):
            s += 'A[%d] = %.7g;\n' % (ssA[i, j], Anum[i,j])

        d['A'] = s

        s = ''
        for i in range(n):
            s += 'b[%d] = %.7g;\n' % (i, bnum[i])

        d['b'] = s

        writecode(self.outdir, self.skel, 'chol_data.h', d)

    def backsubsdata(self, L, ss, b):
        # jem: temporary function.

        d = {}
        d['header'] = header()

        (dss, K) = ss
        b = param('b', value=b)

        # Reverse the role of the keys and values.
        dss2 = {}
        for k in dss.keys():
            dss2[dss[k]] = k

        s = ''
        for k in range(K):
            s += 'L[%d] = %.7g;\n' % (k, L[dss2[k]])
        d['L'] = s

        d['b'] = self.declarematrix(b)

        writecode(self.outdir, self.skel, 'backsubs_data.h', d)

    def multstuff(self, A, AT, m, n):
        """Multiplies and stuffs matrices."""

        # Make a sparse mask of rn. Use A*A'.
        Am = zeros(m, n)
        for k in A.keys():
            Am[k] = 1
        Am = Am*tp(Am)

        p = order(Am)
        Ao = Am[p,p]

        # Design a simple storage scheme for Ao.
        k = 0
        ss = {}
        for (i, j) in zip(Ao.I, Ao.J):
            if i >= j:
                ss[i,j] = k
                k += 1

        # Find reduced Newton system matrix ASAT.
        d = {}
        ASAT = {}
        # Write some code.
        h = param('CM_h', m, 1)
        for i in range(m):
            ip = p[i]
            for j in range(i + 1):
                jp = p[j]
                for k in range(n):
                    if (ip,k) in A and (k,jp) in AT:
                        if ss[i,j] not in d:
                            d[ss[i,j]] = exprtoC(A[ip,k] * h[k] * AT[k,jp])
                        else:
                            d[ss[i,j]] += ' + ' + exprtoC(A[ip,k] * h[k] * AT[k,jp])

        s = ''
        for k in sorted(d.keys()):
            s += 'CM_ASAT[%d] = %s;\n' % (k, d[k])

        return (Ao, s, ss, p)

    def sparsematrixvec(self, ss, m, n, A, x, b):
        """Provides explicit C code for multiplying A*x, where ss is the
        storage scheme for a sparse matrix A, and x and b are both strings."""
        A = nzentries(A)

        x = optvar(x, n)

        for i in range(m):
            d = 0
            for j in range(n):
                if (i, j) in ss:
                    d = d + A[i,j]*x[j]
                    print A[i,j]
                    print d
            print '%s[%d] = %s;' % (b, i, exprtoC(d))

    def getss(self, A):
        # Design a simple storage scheme for Ao.
        m, n = value(size(A))
        A = nzentries(A)

        # Make a sparse mask of A.
        Am = zeros(m, n)
        for k in A.keys():
            Am[k] = 1

        k = 0
        ss = {}
        for (i, j) in zip(Am.I, Am.J):
            if i >= j:
                ss[i,j] = k
                k += 1

        return ss

def withheading(s):
    return s.rstrip() + '\n' + '-'*len(s.rstrip()) + '\n'

def header():
    s = '// CVXMOD code generation.\n'
    s += '// Produced by CVXMOD, %s.\n' % strftime("%Y-%m-%d %H:%M:%S")
    s += '// CXVMOD is Copyright (C) 2006-2008 Jacob Mattingley.\n'
    return s

def writecode(outdir, skeleton, inname, d, outname=None):
    if outname is None:
        outname = inname

    infile = open(os.path.join(skeleton, inname))
    outname = os.path.join(outdir, outname)
    outfile = open(outname, 'w')
    print 'Writing to %s.' % outname
    ril = re.compile('(CVXMOD_IL_(\\w+))')
    rwl = re.compile('(\s*)CVXMOD_WL_(\\w+);?')
    ind = ''
    for l in infile:
        # Test for inline tags. Only one allowed max, per line.
        g = ril.search(l)
        # Test for whole line tags. These take on the same indent as the tag.
        h = rwl.search(l)

        if g:
            try:
                outfile.write(l[:g.start(1)] + str(d[g.group(2)]) + l[g.end(1):])
            except KeyError:
                Jwarn('could not match tag: ' + g.group(2))
                outfile.write(l[:-1] + ' // XXX failed to match tag.\n')
        elif h:
            try:
                for l2 in d[h.group(2)].splitlines():
                    outfile.write(h.group(1) + l2 + '\n')
            except KeyError:
                Jwarn('could not match tag: ' + h.group(2))
                outfile.write(l[:-1] + ' // XXX failed to match tag.\n')
        else:
            outfile.write(ind + l)

    infile.close()
    outfile.close()




#!/usr/bin/env/python

""" MOMENT: Modular Metabolic Engineering Tool 

This script provides a modular metabolic engineering approach
using MOMA analysis. It iterates through a set of given reactions 
and choose the best one for product yield enhancement based on
MOMA flux calculation. Afterwards the reaction is added to the 
set of optimal reactions and the calculation goes on.

Runtime for a set of seven reactions: ~7 hours

Created on 23.09.2016

@author: Julia Helmecke
@email: julia.helmecke@tu-bs.de
"""

from metano.metabolicmodel import MetabolicModel
from metano.paramparser import ParamParser
from metano.defines import FbaParam
from metano.fba import FbAnalyzer
from metano.moma import MomaAnalyzer
from metano.metabolicflux import MetabolicFlux
from metano.moma import main
from multiprocessing import Pool
import os, time, glob, optparse, sys, json, errno


version_number = "alpha_1.3"
""" VERSION HISTORY
    alpha_1.0   basic MOMENT algorithm
    alpha_1.1   added classes and command line tool
    alpha_1.2   added json config file with parser, some small enhancements
    alpha_1.3   added the automatical identification of reaction candidates
    #alpha_1.4   added the possibility to take more than one reaction to the next iteration
"""

# ============================ SETTING PARAMETERS =============================== # 

def unwrap_self_metabolic_engineering(arg, **kwarg):
    return runMoment.metabolic_engineering(*arg, **kwarg)


class OptionParser(optparse.OptionParser):
    def check_required(self, opt):
        option = self.get_option(opt)
        # Assumes the options default is set to None!
        if getattr(self.values, option.dest) is None:
            self.error("%s option not supplied" % option)
            sys.exit()

 


class runMoment():
    
    beginTime = time.time()

    def __init__(self, reactionFile, scenarioFile, outputFile, config, debugging=False): 
        """
        initialize all required values for MOMENT
        """  
        self.config = config
       # define the path to the reaction file:
        self.reactionFile = reactionFile
        # define the path to the scenario file:
        self.scenarioFile = scenarioFile
        # define the path to the output file:
        self.outputFile = outputFile
        # define the path to the output file for all best reactions:
        self.maxkeyOutputFile = outputFile[0:-4] + "_maxkey.txt"
        # define the path to the result folder:
        self.resultFolder = config["resultFolder"]
        # toggle debug mode, if true FBA is performed instead of MOMA:
        self.debugging = debugging
        self.numberMaxThreads = config["numberMaxThreads"]
        # define how many iteration cycles should be calculated:
        self.numberMetabolicInterventions = config["numberMetabolicInterventions"]
        # define how many reactions are considered as best for next cycle:
        #self.numberNextTurnReactions = config["numberNextTurnReactions"]
        # define the flux to be optimized and the wild type target flux:
        self.optimizationFlux = config["targetFlux"]
        # 
        self.transporterPrefix = config["transporterPrefix"]
        # 
        self.ignoreReactionPrefix = config["ignoreReactionPrefix"]
        self.treshold = config["minDiffTreshold"]
        # Set of reactions that are set as already optimized:
        self.setOfInterventions = config["setOfInterventions"]

        self.wtFbaFlux = self.Fba_solution(self.scenarioFile)

        
        # clear all previous parameters and result files and create if not exist:
        try:
            os.makedirs(self.resultFolder)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
            for element in glob.glob(self.resultFolder + "*.txt"):
                os.remove(element)
        try:
            os.makedirs("insce/")
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
            for element in glob.glob("insce/*.txt"):
                os.remove(element)

        # read out the scenario file:
        cfg = open(self.scenarioFile, "r")
        self.insce = ""
        for line in cfg:
            if line[0] not in ("\n"):
                self.insce += line
        cfg.close()


    def Fba_solution(self, sceFile):
        """
        Precalculate the wild type solution for MOMA analysis
        """
        # read out the reactionsfile and parse the metabolic model:
        model = MetabolicModel()
        model.addReactionsFromFile(self.reactionFile)
        pparser = ParamParser()

        # set constraints based on modified scenario file:
        maxmin, objStr, solver, numIter, lb , ub = pparser.parse(sceFile)
        fbaParams = FbaParam(solver, maxmin, objStr, numIter)
        fbaParams.setLinConstraints(pparser.lin_constraints)

        scenarioModel = MetabolicModel()
        scenarioModel.addReactions(model.reactions)
        scenarioModel.setFiniteBounds(lb, ub, False)

        # perform FBA on the model with given constraints:
        fba = FbAnalyzer()
        obj_value, flux, _ = fba.runOnModel(scenarioModel, fbaParams, rmDeadEnds=True)
        return flux


    def identify_candidates(self):
        """
        identify all reaction candidates automatically^or load from file
        """

        # load input file which includes the set of reactions to be modified with
        # their values, seperated by colons:
        self.reactionsDct = {}
        self.setOfCandidates = []
        if len(self.config["setOfReactionCandidates"]) > 0:
            for item in self.config["setOfReactionCandidates"]:
                self.setOfCandidates.append(str(item.split(":")[0].strip()))
                self.reactionsDct[str(item.split(":")[0].strip())] = float(item.split(":")[-1].strip())
        else:
        # 1. perform FBA with optimizationFlux: 
            subst =  "OBJ MAX " + self.optimizationFlux + " \n"
            
            # edit scenario file and write changes to new file:
            with open("sce.txt") as cfg:
                for line in cfg:
                    if line[0] == "#":
                        continue
                    if "OBJ " not in line:
                        subst += line
            scenario = open("insce/optimizationTarget.txt", "w")
            scenario.write(subst)
            scenario.close()
            self.optFbaFlux = self.Fba_solution("insce/optimizationTarget.txt")

            # 2. compare targetFBA with WT FBA:
            comparisonDict = {}
            for rea in self.wtFbaFlux: 
                a = self.optFbaFlux[rea]
                b = self.wtFbaFlux[rea]
                diff = max(a, b) - min(a, b)
                comparisonDict[rea] = diff

            # 3. remove all specific reactions (diff < Th; transporter; production pathway)
            delete = []
            for rea in comparisonDict:
                if comparisonDict[rea] < self.treshold:
                    delete.append(rea)
                elif self.transporterPrefix in rea:
                    delete.append(rea)
                else:
                    for prefix in self.ignoreReactionPrefix:
                        if prefix in rea:
                            delete.append(rea)

            for rea in delete:
                del comparisonDict[rea]
            # check how many candidate reactions were calculated and
            # give the possibility to abort and adjust treshold
            if len(comparisonDict) > 30 or len(comparisonDict) < 5:
                print """WARNING!! %d candidate reactions were calculated !\n
                Try to adjust the treshold in the config file!""" % len(comparisonDict)
                answer = raw_input("Do you want to continue? (y/n)\n")
                if answer == "n":
                    exit()

            # 4. export values into dict:
            for rea in comparisonDict:
                self.setOfCandidates.append(rea)
                self.reactionsDct[rea] = float(self.optFbaFlux[rea])

        for element in self.setOfInterventions:
            self.setOfCandidates.remove(element) 
        print "Candidate Reactions:", self.setOfCandidates

          

    def metabolic_engineering(self,argument):
        """
        calculate the MOMA result for target flux while setting constraints
        to all optimization reactions.
        """
        new_rea, setOfInterventions_list = argument
        name = str(len(setOfInterventions_list)) + "_" + new_rea
        
        if new_rea == []:
            pass
        else:
            # modify the scenario file by setting upper and lower bounds
            # for each manipulated reaction:
            subst =  "LB " + new_rea + " " + str(self.reactionsDct[new_rea]) + " \n"
            subst += "UB " + new_rea + " " + str(self.reactionsDct[new_rea]) + " \n\n"
            for element in setOfInterventions_list:
                subst += "LB " + element + " " + str(self.reactionsDct[element]) + " \n"
                subst += "UB " + element + " " + str(self.reactionsDct[element]) + " \n"
            
            # write changes to scenario file:
            cfg = open("insce/insce_" + name + ".txt", "w")
            cfg.write(self.insce + "\n\n")
            cfg.write(subst)
            cfg.close()

            # read out the reactionsfile and parse the metabolic model:
            model = MetabolicModel()
            model.addReactionsFromFile(self.reactionFile)
            pparser = ParamParser()

            # set constraints based on modified scenario file:
            maxmin, objStr, solver, numIter, lb , ub = pparser.parse("insce/insce_" + name + ".txt")
            fbaParams = FbaParam(solver, maxmin, objStr, numIter)
            fbaParams.setLinConstraints(pparser.lin_constraints)
            scenarioModel = MetabolicModel()
            scenarioModel.addReactions(model.reactions)
            scenarioModel.setFiniteBounds(lb, ub, False)

            #save flux distribution of WT FBA into variable for usage during MOMA analysis:
            

            # during debugging, only FBA is performed:
            if self.debugging == True:
                solutionFlux = self.Fba_solution("sce.txt")
            else:
                # perform MOMA analysis and save the resulting flux 
                # distribution in "solutionFlux":
                moma = MomaAnalyzer("default")
                dist, solutionFlux, status = moma.runOnModel(scenarioModel, 
                    self.wtFbaFlux, fbaParams.linConstraints, numIter)[:3]
            

            # save the flux through optimization flux
            try:
                targetFlux = solutionFlux.fluxDict[self.optimizationFlux]
            except:
                targetFlux = 0
                print "Warning:", self.optimizationFlux, "could not be found!"
                
            # write result to file. This step is necessary when 
            # using multiprocessing. In this way the cores dont
            # get in each others way:
            maxFluxFile = open(self.resultFolder + name + ".txt", "w")
            output = new_rea + " " + str(targetFlux) + "\n"
            maxFluxFile.write(output)

            # delete the scenario file:
            #os.remove("self.insce/self.insce_" + name + ".txt")
            return targetFlux


    def moment(self):
        
        # create empty output file:
        maxkey_output = ""
        text_file = open(self.outputFile, "w")
        text_file.write("Number;Reaction;Target_flux\n")
        text_file.close()

        # set number of cores to be used for multiprocessing:
        pool=Pool(processes=self.numberMaxThreads)

        # perform calculation of the iteration cycle:
        while len(self.setOfInterventions) < self.numberMetabolicInterventions:
            # set the calculation parameter and perform MOMA analysis:
            maxFluxDict = {}
            modify = []
            for entry in self.setOfCandidates:
                modify.append((entry, self.setOfInterventions))
            
            if self.debugging == True:
                pool=Pool(processes=1)
            pool.map(unwrap_self_metabolic_engineering, zip([self]*len(modify), modify))
            #pool.map(self.metabolic_engineering, modify)
            # read out all results from primary files:
            for element in self.setOfCandidates: 
                with open(self.resultFolder + str(len(self.setOfInterventions)) + "_" + element + ".txt") as f:
                    for line in f:
                        print line
                        maxFluxDict[line.split()[0]] = float(line.strip("\n").split()[-1])
            
            # write all results to output file:
            text_file = open(self.outputFile, "a")
            output = ""
            for key in maxFluxDict:
                output += str(len(self.setOfInterventions)) + ";" + key + ";" + str(maxFluxDict[key]) + "\n"
            text_file.write(output)
            text_file.close()
            print output
            
            # save the best result into self.setOfInterventions list and delete from new_rea:
            print max(maxFluxDict, key=maxFluxDict.get)
            maxkey = max(maxFluxDict, key=maxFluxDict.get)
            print "***   ", maxkey, "=", maxFluxDict[maxkey], "   ***\n\n\n"
            self.setOfInterventions.append(maxkey)
            self.setOfCandidates.remove(maxkey)
            maxkey_output += maxkey + "; " + str(maxFluxDict[maxkey]) + "\n"
            
            print "Runtime for", len(self.setOfInterventions), "reaction(s):\n", \
                  round((time.time() - self.beginTime)/60, 2), "minutes\n"


        # finally write the set of best reactions to output file:
        text_file = open(self.maxkeyOutputFile, "w")
        text_file.write(maxkey_output)
        text_file.close()
        print "\n\nMAXKEY OUTPUT:\n", maxkey_output

        print "RUNTIME:\n", \
              round(time.time() - self.beginTime, 2), "seconds\n", \
              round((time.time() - self.beginTime)/60, 2), "minutes\n", \
              round((time.time() - self.beginTime)/60/60, 2), "hours\n"


def main():
    usage = "Usage: %prog [options]"
    version = "Modular Metabolic Engineering Test \nVersion %s" % version_number 
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform Flux Balance Analysis on the network given "
                           "by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="scenarioFile",
                      help="use the given WT scenario FILE for Flux Balance "
                           "Analysis", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write output of MOMENT to FILE",
                      metavar="FILE")
    parser.add_option("-c", "--config", dest="configFile",
                      help="read all configurations from JSON FILE",
                      metavar="FILE")
    #parser.add_option("-t", "--target", action="store", type="string",
    #                  dest="targetFlux", help="target flux that should be maximized")
    parser.add_option("-D", "--debug", action="store_true",
                      dest="debugging", help="uses FBA instead of MOMA for shorter Runtime")
    #parser.add_option("-N", "--number", action="store", type="int",
    #                  dest="numberMetabolicInterventions", help="number of modified reactions")
    parser.set_defaults(debugging=False)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-o")
    parser.check_required("-c")

    with open(options.configFile) as cf:
        cfDict = json.load(cf)
    
    myMoment = runMoment(options.reactionFile,
        options.scenarioFile,
        options.outputFile,
        cfDict,
        options.debugging)

    myMoment.identify_candidates()
    myMoment.moment() 
    
if __name__ == "__main__":
    main()




"""
Convex optimization modeling for cvxopt.

"""

# Copyright (C) 2006-2008 Jacob Mattingley and Stephen Boyd.
#
# This file is part of CVXMOD.
#
# CVXMOD is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# CVXMOD is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import cvxmod
from cvxmod.base import *
from cvxmod.base import addormultfunction, withbrackets
from cvxmod.util import iterable
from cvxmod.errors import *

__all__ = ["eyesymbol", "constsymb", "equivdims", "compactdims", "ch", "cv",
           "stdformlinsymb", "strsymbmatrix"]

class eyesymbol(symbol):
    symm = True
    psd = True
    def __init__(self, size):
        self.rows = size
        self.cols = size

    def __repr__(self):
        return "<%sx%s identity matrix>" % (withbrackets(self.rows),
           withbrackets(self.cols))

    def __str__(self):
        return 'I(%s)' % rows(self)

    def _getvalue(self):
        # jem. make this cache.
        return speye(value(self.rows))
    value = property(_getvalue)

    def nnz(self):
        return self.rows

    def nzentries(self):
        d = {}
        for i in range(value(self.rows)):
            d[i,i] = 1

        return d

class constsymb(symbol):
    def __init__(self, rs, cs, constval):
        self.rows = rs
        self.cols = cs
        self.role = 'constant'
        self.constval = constval

    def __repr__(self):
        if self.constval == 1:
            descr = "ones"
        else:
            descr = "zeros"
        descr += " "

        if equiv(self.cols, 1):
            descr += "vector"
        else:
            descr += "matrix"
        return "<%sx%s all %s>" % (withbrackets(self.rows),
           withbrackets(self.cols), descr)

    def __str__(self):
        if self.constval == 1:
            descr = "ones"
        else:
            descr = "zeros"

        return "%s(%s, %s)" % (descr, str(withbrackets(self.rows)),
                               str(withbrackets(self.cols)))

    def _getvalue(self):
        if self.constval == 1:
            return ones(value(self.rows), value(self.cols))
        else:
            return zeros(value(self.rows), value(self.cols))
    value = property(_getvalue)

    def etranspose(self):
        return constsymb(self.cols, self.rows, self.constval)

    def nnz(self):
        if iszero(self):
            return 0
        else:
            return self.rows*self.cols

    def nzentries(self):
        if iszero(self):
            return {}
        else:
            d = {}
            for i in range(value(self.rows)):
                for j in range(value(self.cols)):
                    d[i,j] = self.constval

            return d
    
def equivdims(lhs, rhs):
    """Compare two representations of a dimension, and see if they are the
    same."""

    if lhs is rhs:
        return True

    if lhs - rhs is 0:
        return True

    l = lhs - rhs

    if l is 0:
        return True
    elif isinstance(l, int):
        return False

    if getdimmult(l, None) is not 0:
        return False

    for p in getdims(l):
        a = getdimmult(l, p)
        if iszero(a):
            continue
        else:
            return False

    return True

def compactdims(x):
    """Turn b + b into 2*b."""
    # jem: will fail with products of dims.
    y = 0
    for p in sorted(getdims(x), lambda x, y: cmp(str(x).lower(),
                                                   str(y).lower())):
        y = y + getdimmult(x, p)*p

    return y + getdimmult(x, None)

class concathorizsymb(addormultfunction):
    def __init__(self, lhs, rhs):
        self.rows = rows(lhs)
        self.cols = compactdims(cols(lhs) + cols(rhs))

        self.lhs = lhs
        self.rhs = rhs

    def __repr__(self):
        return "<%sx%s horiz concat [%s; %s]>" % \
                (withbrackets(rows(self)), withbrackets(cols(self)),
                 str(self.lhs), str(self.rhs))

    def __str__(self):
        if isinstance(self.lhs, concathorizsymb):
            if isinstance(self.rhs, concathorizsymb):
                # both.
                return "%s %s" % (str(self.lhs)[:-1], str(self.rhs)[1:])
            else:
                # lhs only.
                return "%s %s]" % (str(self.lhs)[:-1], str(self.rhs))
        elif isinstance(self.rhs, concathorizsymb):
            # rhs only.
            return "[%s %s" % (str(self.lhs), str(self.rhs)[1:])
        else:
            return "[%s %s]" % (str(self.lhs), str(self.rhs))

    def _getvalue(self):
        return concathoriz(value(self.lhs), value(self.rhs))
    value = property(_getvalue)

    def getvarmult(self, var):
        l = getvarmult(self.lhs, var)
        r = getvarmult(self.rhs, var)
        if iszero(l):
            if var is None:
                l = zeros(size(self.lhs), symb=True)
            else:
                l = zeros(rows(self.lhs), rows(var), symb=True)
        if iszero(r):
            if var is None:
                r = zeros(size(self.rhs), symb=True)
            else:
                r = zeros(rows(self.rhs), rows(var), symb=True)

        return ch(l, r)

    def getvecmult(self, var):
        raise NotImplementedError('getvecmult for concathorizsymb')

    def nnz(self):
        return nnz(self.lhs) + nnz(self.rhs)

    def nzentries(self):
        d = {}
        d = nzentries(self.lhs)
        drhs = nzentries(self.rhs)

        n = value(cols(self.lhs))
        for (i, j) in drhs.keys():
            d[i,j+n] = drhs[i,j]

        return d

class concatvertsymb(addormultfunction):
    def __init__(self, lhs, rhs):
        self.cols = cols(lhs)
        self.rows = compactdims(rows(lhs) + rows(rhs))

        self.lhs = lhs
        self.rhs = rhs

    def __repr__(self):
        return "<%sx%s vert concat [%s; %s]>" % \
                (withbrackets(rows(self)), withbrackets(cols(self)),
                 str(self.lhs), str(self.rhs))

    def __str__(self):
        if isinstance(self.lhs, concatvertsymb):
            if isinstance(self.rhs, concatvertsymb):
                # both.
                return "%s %s" % (str(self.lhs)[:-1], str(self.rhs)[1:])
            else:
                # lhs only.
                return "%s; %s]" % (str(self.lhs)[:-1], str(self.rhs))
        elif isinstance(self.rhs, concatvertsymb):
            # rhs only.
            return "[%s; %s" % (str(self.lhs), str(self.rhs)[1:])
        else:
            return "[%s; %s]" % (str(self.lhs), str(self.rhs))
    def _getvalue(self):
        return concatvert(value(self.lhs), value(self.rhs))
    value = property(_getvalue)

    def getvarmult(self, var):
        l = getvarmult(self.lhs, var)
        r = getvarmult(self.rhs, var)
        if iszero(l):
            if var is None:
                l = zeros(size(self.lhs), symb=True)
            else:
                l = zeros(rows(self.lhs), rows(var), symb=True)
        if iszero(r):
            if var is None:
                r = zeros(size(self.rhs), symb=True)
            else:
                r = zeros(rows(self.rhs), rows(var), symb=True)

        return cv(l, r)

    def getvecmult(self, var):
        raise NotImplementedError('getvecmult for concatvertsymb')

    def nnz(self):
        return nnz(self.lhs) + nnz(self.rhs)
    def nzentries(self):
        d = {}
        d = nzentries(self.lhs)
        drhs = nzentries(self.rhs)

        m = value(rows(self.lhs))
        for (i, j) in drhs.keys():
            d[i+m,j] = drhs[i,j]

        return d

def ch(lhs, rhs):
    """Concatenate horizontally and symbolically, performing any available
    reductions."""
    if lhs is None:
        return rhs
    elif rhs is None:
        return lhs
    elif not equivdims(rows(lhs), rows(rhs)):
        raise DimError

    try:
        if lhs.constval == rhs.constval:
            return constsymb(rows(lhs), compactdims(cols(lhs) + cols(rhs)),
                             lhs.constval)
    except AttributeError:
        pass

    if isinstance(lhs, concathorizsymb):
        if isinstance(rhs, concathorizsymb):
            try:
                if lhs.rhs.constval == rhs.lhs.constval:
                    c = constsymb(rows(lhs.rhs),
                                  compactdims(cols(lhs.rhs) + cols(rhs.lhs)),
                                  lhs.rhs.constval)
                    return ch(ch(lhs.lhs, c), rhs.rhs)
            except AttributeError:
                pass
        else:
            try:
                if lhs.rhs.constval == rhs.constval:
                    c = constsymb(rows(lhs),
                                  compactdims(cols(lhs.rhs) + cols(rhs)),
                                  lhs.rhs.constval)
                    return ch(lhs.lhs, c)
            except AttributeError:
                pass
    elif isinstance(rhs, concathorizsymb):
        try:
            if lhs.constval == rhs.lhs.constval:
                c = constsymb(rows(lhs),
                              compactdims(cols(lhs) + cols(rhs.lhs)),
                              lhs.constval)
                return ch(c, rhs.rhs)
        except AttributeError:
            pass

    return concathorizsymb(lhs, rhs)
    
def cv(lhs, rhs):
    """Concatenate vertically and symbolically, performing any available
    reductions."""
    if lhs is None:
        return rhs
    elif rhs is None:
        return lhs
    elif not equivdims(cols(lhs), cols(rhs)):
        raise DimError

    try:
        if lhs.constval == rhs.constval:
            return constsymb(compactdims(rows(lhs) + rows(rhs)), cols(lhs),
                             lhs.constval)
    except AttributeError:
        pass

    if isinstance(lhs, concatvertsymb):
        if isinstance(rhs, concatvertsymb):
            try:
                if lhs.rhs.constval == rhs.lhs.constval:
                    c = constsymb(compactdims(rows(lhs.rhs) + rows(rhs.lhs)),
                                  cols(lhs.rhs), lhs.rhs.constval)
                    return cv(cv(lhs.lhs, c), rhs.rhs)
            except AttributeError:
                pass
        else:
            try:
                if lhs.rhs.constval == rhs.constval:
                    c = constsymb(compactdims(rows(lhs.rhs) + rows(rhs)),
                                  cols(lhs.rhs), lhs.rhs.constval)
                    return cv(lhs.lhs, c)
            except AttributeError:
                pass
    elif isinstance(rhs, concatvertsymb):
        try:
            if lhs.constval == rhs.lhs.constval:
                c = constsymb(compactdims(rows(lhs) + rows(rhs.lhs)),
                              cols(lhs), lhs.constval)
                return cv(c, rhs.rhs)
        except AttributeError:
            pass

    return concatvertsymb(lhs, rhs)

def stdformlinsymb(objv, lin, optvars, stdform='cAbGh'):
    # Symbolic version of stdformlin.

    A = None
    b = None
    G = None
    h = None
    for d in lin:
        s = d.stdlhs()
        r = None
        if d.sign == '==':
            for v in optvars:
                a = getvarmult(s, v)
                if iszero(a):
                    a = zeros(rows(s), rows(v), symb=True)
                r = ch(r, a)
            A = cv(A, r)
            bb = -getvarmult(s, None)
            if is1x1(bb) and not equiv(size(bb), size(s)):
                bb = bb*ones(size(s), symb=True)
            b = cv(b, bb)
        else:
            for v in optvars:
                a = getvarmult(s, v)
                if iszero(a):
                    a = zeros(rows(s), rows(v), symb=True)
                r = ch(r, a)
            G = cv(G, r)
            bb = -getvarmult(s, None)
            if is1x1(bb) and not equiv(size(bb), size(s)):
                if iszero(bb):
                    bb = zeros(size(s), symb=True)
                else:
                    bb = bb*ones(size(s), symb=True)
            h = cv(h, bb)

    if objv is None and A is not None:
        c = zeros(cols(A), 1, symb=True)
    else:
        c = None
        for v in optvars:
            a = etp(getvarmult(objv, v))
            if iszero(a):
                a = zeros(rows(v), 1, symb=True)
            c = cv(c, a)

    if stdform == 'cAbGh':
        return ((c, A, b, G, h), optvars)
    elif stdform == 'cAbp':
        m = rows(A)
        n = cols(A)
        p = rows(G)
        ysf = optvar('ysf', p)
        optvars.append(ysf)
        
        # Define new standard form as follows:
        # minimize ct^T*xt, subject to At*xt == bt, St*xt >= 0.
        # ct = [0; c]
        # At = [0 A; I G]
        # bt = [b; h]
        # xt = [x; ysf]
        # p = p.
        ct = cv(zeros(p, 1, True), c)
        At = cv(ch(zeros(m, p, True), A), ch(eye(p, True), G))
        bt = cv(b, h)

        return ((ct, At, bt, p), optvars)

def strsymbmatrix(A, name):
    s = ''
    s += 'rows: %s.\n' % withbrackets(rows(A))
    s += 'cols: %s.\n' % withbrackets(cols(A))
    s += '%s =\n' % name
    s += str(A).replace(';', ';\n') + '\n'
    return s



#!/usr/bin/env python

""" Script for converting a reaction file to a stoichiometric matrix

This script constructs a ReactionParser and calls its exportMatrixAsDotM()
method, which exports the stoichiometric matrix as a .m file.


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

from fba import OptionParser
from metabolicmodel import MetabolicModel
from defines import exportMatrixAsDotM, COPYRIGHT_VERSION_STRING
import os


def main():
    # Parse command line

    usage = "Usage: %prog [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile", help="use the "
                      "given reaction FILE as input", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile", help="write the "
                      "stoichiometric matrix to FILE (should have .m extension)",
                      metavar="FILE")

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-o")

    # Parse reaction file

    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile)
    except IOError, strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print strerror
        exit()
    except SyntaxError, strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print strerror
        exit()

    # Export matrix

    try:
        exportMatrixAsDotM(model.getStoichiometricMatrix(), options.outputFile)
    except IOError, strerror:
        print ("An error occurred while trying to write file %s:" %
               os.path.basename(options.outputFile))
        print strerror
        exit()


if __name__ == "__main__":
    main()



"""
This module defines the class NonLinearProblem, which represents a nonlinear
problem to be solved by an appropriate solver.


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

from openopt import NLP
from linearproblem import LinearProblem


class NonLinearProblem(LinearProblem):

    # Table of nonlinear problem solvers with corresponding OpenOpt name
    solvers = {'default'  : 'algencan',
               'ralg'     : 'ralg',           # for medium-sized problems;
                                              #  ill-conditioned
               'cobyla'   : 'scipy_cobyla',   # recommended
               'algencan' : 'algencan',       # recommended
               'slsqp'    : 'scipy_slsqp',
               'ipopt'    : 'ipopt',
               'lincher'  : 'lincher'}        # very primitive


    def __init__(self, Aeq, beq=None, Aineq=None, bineq=None, lb=None, ub=None,
                 solver=solvers['default']):
        """ initialize the nonlinear problem

        Keyword arguments:

        Aeq     -- matrix of equality constrains (Aeq * v = beq)
        beq     -- right-hand side vector of equality constraints
        Aineq   -- matrix of inequality constrains (Aineq * v <= bineq)
        bineq   -- right-hand side vector of inequality constraints
        lb      -- list of lower bounds (indexed like matrix columns)
        ub      -- list of upper bounds (indexed like matrix columns)
        solver  -- solver to be used for Nonlinear Programming
        """
        LinearProblem.__init__(self, Aeq, beq, Aineq, bineq, lb, ub, solver)
        if solver.lower() in LinearProblem.solvers:
            # If problem has been raised from a linear problem, use default
            # solver.
            self.solver = NonLinearProblem.solvers["default"]
        self.obj      = None
        self.d_obj    = None
        self.nlc      = None
        self.d_nlc    = None
        self.x0       = [0.]*len(lb)
        self.iterator = None
        self.rlist    = None


    def setObjective(self, f):
        self.obj = f

    def resetObjective(self):
        """ shadow LinearProblem.resetObjective """
        pass

    def setObjGrad(self, df):
        self.d_obj = df

    def setNonlinearConstraints(self, nlc):
        self.nlc = nlc

    def setNlcGrad(self, d_nlc):
        self.d_nlc = d_nlc


    def setStartPoint(self, x0):
        if len(x0) != len(self.lb):
            raise ValueError("Error: Starting point has wrong dimension.")
        self.x0 = x0
        self.rlist = None

    def setStartPointIterator(self, iterator):
        self.iterator = iterator


    def minimize(self, **kwargs):
        """ solve the nonlinear problem using OpenOpt

        Returns:
        obj_value, solution

        obj_value -- value of the objective function at the discovered solution
        solution  -- the solution flux vector (indexed like matrix columns)
        """
        if self.iterator is None:
            nlp = NLP(self.obj, self.x0, df=self.d_obj, c=self.nlc,
                      dc=self.d_nlc, A=self.Aineq, Aeq=self.Aeq, b=self.bineq,
                      beq=self.beq, lb=self.lb, ub=self.ub, **kwargs)
            nlp.debug=1
            nlp.plot = False
            nlp.checkdf()
            if self.nlc is not None:
                nlp.checkdc()

            r = nlp.solve(self.solver)

        else:
            self.rlist = []
            for x0 in self.iterator:
                nlp = NLP(self.obj, x0, df=self.d_obj, c=self.nlc,
                          dc=self.d_nlc, A=self.Aineq, Aeq=self.Aeq,
                          b=self.bineq, beq=self.beq, lb=self.lb, ub=self.ub,
                          **kwargs)
                r = nlp.solve(self.solver)
                if r.istop > 0 and r.ff == r.ff:
                    self.rlist.append(r)

            if self.rlist != []:
                r = min(self.rlist, key=lambda x : x.ff)

        if r.istop <= 0 or r.ff != r.ff:  # check halting condition
            self.obj_value = 0.
            self.solution = []
            self.istop = r.istop
        else:
            self.obj_value = r.ff
            self.solution = r.xf
            self.istop = r.istop

        return self.obj_value, self.solution


    def getHistogram(self, numBins=-1):
        """ compute a histogram over all found solutions (objective function
            values)

            Returns:
            histogram, minval, maxval
        """
        numPoints = len(self.rlist)
        if numPoints == 0:
            return [], 0., 0.

        if numBins == 0:
            raise ValueError("numBins must not be zero")

        if numBins < 0:
            numBins = min(int(round(numPoints / 10.)), 1)

        minval = min(self.rlist, key=lambda x : x.ff).ff
        maxval = max(self.rlist, key=lambda x : x.ff).ff
        if minval == maxval:
            return [numPoints], minval, maxval
        binWidth = (maxval-minval)/float(numBins)

        histogram = [0]*(numBins+1) # Extra bin for maxval

        for r in self.rlist:
            histogram[int((r.ff-minval)/binWidth)] += 1

        histogram[-2] += histogram[-1]
        del histogram[-1]

        return histogram, minval, maxval


    def eval(self, v):
        """ evaluate objective function for vector v
        """
        return self.d_obj(v)



from setuptools import setup
from setuptools.command import sdist

sdist.walk_revctrl = lambda **kwargs: []

setup(
    name = "metano",
    version = "1.1.0",
    description = "metano - A Toolkit for Metabolic Network Analysis and "
                  "Optimization",
    author = "Alexander Riemer",
    author_email = "al.riemer@tu-bs.de",
    url='http://metano.tu-bs.de/',
    packages = ["metano", "cvxmod"],
    package_dir = {"metano" : "src", "cvxmod" : "cvxmod"},
    package_data={
        "metano" : ["example/*.txt", "example/*.sh"]
        },
    entry_points = {
      "console_scripts": [
         'deadends.py = metano.deadends:main',
         'fba.py = metano.fba:main',
         'fva.py = metano.fva:main',
         'moma.py = metano.moma:main',
         'knockout.py = metano.knockout:main',
         'sbmlparser.py = metano.sbmlparser:main',
         'sbmlwriter.py = metano.sbmlwriter:main',
         'splitratios.py = metano.splitratios:main',
         'mfm.py = metano.mfm:main',
         'modelassert.py = metano.modelassert:main',
         'moment.py = metano.moment:main',
         'to_check_constraints.py = metano.to_check_constraints:main',
         'to_diff_solutions.py = metano.to_diff_solutions:main',
         'to_fix_names.py = metano.to_fix_names:main',
         'to_list_active_metab.py = metano.to_list_active_metab:main',
         'to_rea2m.py = metano.to_rea2m:main',
      ]
    },
    classifiers = [
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
# Development Status :: 2 - Pre-Alpha
# Development Status :: 3 - Alpha
# Development Status :: 4 - Beta
# Development Status :: 5 - Production/Stable
# Development Status :: 6 - Mature
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords = "genome-scale metabolic models flux balance analysis",
    license = "GPL",
    install_requires = [
        "numpy",
        "swiglpk",
        "pymprog>=1.0",
        "ecos",
        "scipy",
        "openopt",
        "cvxpy",
        "cvxopt<=1.1.8",
]
    )


"""
This module defines the class LinearConstraint, which represents a linear
equality or inequality constraint.


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

class LinearConstraint(object):
    """ Represents a single linear equality or inequality constraint

    Member variables:

    lhs  -- list of (coef, name) tuples
    rhs  -- value of right-hand side of equality/inequality (float constant)
    isEq -- if True, interpretation is 'lhs = rhs', else 'lhs < rhs'
    """

    def __init__(self, lhs, sign=None, rhs=None):
        """ create a new linear constraint

        Keyword arguments:

        lhs  -- either a list of (coef, reaction name) tuples or a list of such
                tuples followed by a sign (see below) and a float value -
                explicitly given sign and rhs take precedence over those
                potentially defined in lhs
        sign -- comparison sign (<, >, =, <=, >=, or ==)
        rhs  -- value of right-hand side of equality/inequality (float constant)
        
        Sign and rhs are optional, defaults are '<' and 0.
        """
        self.clear()
        lhsExtra = []
        for elem in lhs:
            try:
                _, _ = elem
                self.lhs.append(elem)
            except (ValueError, TypeError):
                lhsExtra.append(elem)

        if lhsExtra:
            if not sign:
                sign = lhsExtra[0]
            if len(lhsExtra) > 1:
                if not rhs:
                    rhs = lhsExtra[1]

        if rhs:
            self.rhs = rhs

        # Parse sign
        if sign:
            sign = sign[0]
            if sign == '=':
                self.isEq = True
    
            elif sign == '>':
                # Flip inequality by negating both signs
                self.rhs = -self.rhs
                self.lhs = [(-coef, name) for coef, name in self.lhs]


    def clear(self):
        self.lhs = []
        self.rhs = 0.
        self.isEq = False


    def getSign(self):
        return '=' if self.isEq else '<'


    def __str__(self):
        s = ""
        for coef, name in self.lhs:
            if not s:
                s = "%g %s" % (coef, name)
            else:
                if coef < 0.:
                    s+= " - %g %s" % (-coef, name)
                else:
                    s+= " + %g %s" % (coef, name)
        return "%s %s %g" % (s, self.getSign(), self.rhs)


    def __repr__(self):
        return "LinearConstraint(%r, %r, %r)" % (self.lhs, self.getSign(),
                                                 self.rhs)



#!/usr/bin/env python

""" metano Knockout Analysis

This script performs automatic knockout analysis using Minimization of Metabolic
Adjustment (MOMA) on the metabolic network (given as a set of reaction
equations) and inequality constraints. This is done by sequentially setting the
fluxes through given groups of reactions to zero and performing MOMA against the
wildtype.


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

# TODO Release:
# - remove wMOMA: _DEFAULT_ALPHA/BETA, weighted, weights, weightVec
# - remove option_group("Extra options for weighted MOMA")
# - remove step 7. (Optionally) Read FVA solution and compute weights

from fba import OptionParser, FbAnalyzer
from moma import MomaAnalyzer, _DEFAULT_ALPHA, _DEFAULT_BETA
from fva import FvAnalyzer
from reactionparser import ReactionParser
from paramparser import ParamParser
from metabolicmodel import MetabolicModel
from metabolicflux import MetabolicFlux
from defines import (SolverStatus, FbaParam, padNumber, When, decodeWhen,
                     COPYRIGHT_VERSION_STRING)
from numpy import array, dot, nan, inf, empty
import csv, os
try:
    from mpi4py import MPI
    _mpi_avail = True
except ImportError:
    _mpi_avail = False


def _mpi_exit(comm=None, errno=0):
    if comm:
        comm.Abort(errno)
    else:
        exit(errno)

_CUTOFF_LETHAL = 1E-5
_FILE_SUFFIX = '.txt'


def _makeDefaultKoGroups(model):
    """ generate a list of singleton reaction groups from given MetabolicModel
    """
    return [(rea.name, [rea.name]) for rea in model]


def serial_knockout(model, solver, objective, wtSolution, fbaParams,
                    koGroups=None, filePrefix=None, wtObjVal=None, weights=None,
                    numIter=1, useMoma=When.AUTO,
                    lethalityCutoff=_CUTOFF_LETHAL):
    """ perform knockout analysis in serial fashion (single process)

    Keyword arguments:

    model      -- the MetabolicModel
    solver     -- name of QP/NLP solver (or "default")
    objective  -- coefficient vector of linear objective function of FBA
    wtSolution -- FBA solution for wildtype
    fbaParams  -- parameters for FBA (incl. linear equality and inequality
                                      constraints)
    koGroups   -- list of pairs (group name, list of reactions)
    filePrefix -- if given, write MOMA/FBA solutions to files starting with this
                  prefix; if not given, solutions are discarded
    wtObjVal   -- objective function value for wildtype solution (optional)
    weights    -- weight vector for weighted MOMA (None -> perform regular MOMA,
                  else: weight flux i with weights[i])
    numIter    -- number of NLP runs (if NLP solver is used) for MOMA
    useMoma    -- selector of optimization strategy, enum values (class When):
                  NEVER  - always perform FBA (LP)
                  AUTO   - perform FBA first, followed by MOMA if not lethal
                  ALWAYS - always perform MOMA (QP)
    lethalityCutoff
               -- threshold for biomass flux signifying lethality

    Returns list of (distance, diff, obj_val) tuples, indexed like koGroups,

    distance -- value of actual MOMA/FBA objective function at solution
    diff     -- summed absolute difference between mutant and wildtype solutions
    obj_val  -- value of FBA objective function at MOMA/FBA solution for mutant
    """
    if not koGroups:
        koGroups = _makeDefaultKoGroups(model)
#    print "serial: %d reactions, %d knockout groups" % (len(model),
#                                                        len(koGroups))
    wtVec = wtSolution.getVecOrderedByModel(model)
    if wtObjVal is None:
        wtObjVal = dot(objective, wtVec)
    moma = MomaAnalyzer(solver)

    result = []
    for group, reaList in koGroups:
        print group

        # Restrict flux through all reactions in group to zero
        tmp_lb = {}  # {index in model : LB value}
        tmp_ub = {}  # {  - '' -       : UB  ''  }
        for rea in reaList:
            try:
                rIndex = model.reactionDict[rea]
            except KeyError:
                # Skip any blocked reactions (knockout has no effect)
                continue
            tmp_lb[rIndex] = model.reactions[rIndex].lb
            tmp_ub[rIndex] = model.reactions[rIndex].ub
            model.reactions[rIndex].lb = model.reactions[rIndex].ub = 0.

        if not tmp_lb:
            print "  -> skipped (only blocked reactions)"
            result.append((0., 0., wtObjVal))
            if filePrefix:
                wtSolution.writeToFile(filePrefix+group+_FILE_SUFFIX)
            continue

        obj_val = lethalityCutoff + 1.

        # Perform FBA first to check for lethality
        if useMoma != When.ALWAYS:
            fba = FbAnalyzer(fbaParams.solver)
            # Note: Model is already reduced
            obj_val, solutionFlux = fba.runOnModel(model, fbaParams,
                                                   rmDeadEnds=False)[:2]
            status = SolverStatus.PREPARED
            dist = None
            if not solutionFlux:
                obj_val = 0.

        # Perform MOMA if knockout not already predicted by FBA to be lethal
        if useMoma != When.NEVER and abs(obj_val) > lethalityCutoff:
            dist, solutionFlux, status = moma.runOnModel(model, wtSolution,
                fbaParams.linConstraints, numIter, weights)[:3]

        if status == SolverStatus.UNKNOWN:
            result.append((nan, nan, 0.))
        elif status in (SolverStatus.PRIM_INFEAS, SolverStatus.DUAL_INFEAS):
            result.append((inf, inf, 0.))
        elif len(solutionFlux) == 0:
            if status == SolverStatus.PREPARED:
                # FBA infeasible
                result.append((inf, inf, 0.))
            else:
                result.append((nan, nan, 0.))
        else:
            diff = sum(solutionFlux.absDiff(wtSolution).fluxDict.values())
            if dist is None:
                dist = moma.evalObjFunc(wtVec,
                    solutionFlux.getVecOrderedByModel(model), weights)
            else:
                obj_val = max(0., dot(objective,
                                      solutionFlux.getVecOrderedByModel(model)))
            result.append((dist, diff, obj_val))
            if filePrefix:
                solutionFlux.writeToFile(filePrefix+group+_FILE_SUFFIX)

        # Restore original constraints
        for rea in reaList:
            try:
                rIndex = model.reactionDict[rea]
            except KeyError:
                continue
            model.reactions[rIndex].lb = tmp_lb[rIndex]
            model.reactions[rIndex].ub = tmp_ub[rIndex]

    return result


def parallel_ko_dispatch(comm, numprocs, model, solver, objective, wtSolution,
                         wtObjVal=None, koGroups=None, filePrefix=None,
                         weights=None):
    """ root process (dispatcher) for parallel knockout analysis

    Keyword arguments:

    comm       -- MPI communicator
    numprocs   -- number of processes (including root process)
    model      -- the MetabolicModel
    solver     -- name of solver (or "default")
    objective  -- coefficient vector of linear objective function of FBA
    wtSolution -- FBA solution for wildtype
    wtObjVal   -- objective function value for wildtype solution (optional)
    koGroups   -- list of pairs (group name, list of reactions)
    filePrefix -- if given, write MOMA/FBA solutions to files starting with this
                  prefix; if empty, solutions are discarded
    weights    -- weight vector for weighted MOMA (None -> perform regular MOMA,
                  else: weight flux i with weights[i])

    Returns list of (distance, diff, obj_val) tuples, indexed like koGroups,

    distance -- value of actual MOMA/FBA objective function at solution
    diff     -- summed absolute difference between mutant and wildtype solutions
    obj_val  -- value of FBA objective function at MOMA/FBA solution for mutant
    """
    if not koGroups:
        koGroups = _makeDefaultKoGroups(model)
#    print "dispatch: %d reactions, %d knockout groups" % (len(model),
#                                                          len(koGroups))
    wtVec = wtSolution.getVecOrderedByModel(model)
    if wtObjVal is None:
        wtObjVal = dot(objective, wtVec)
    moma = MomaAnalyzer(solver)

    nJobs = len(koGroups)
    stopMsg = array(-1, 'i')  # stop message (-1)
    status = MPI.Status()
    result = [None]*nJobs
    jobList = [-1]*numprocs  # mapping of process -> job (jobList[0] is dummy)
    # receive buffer for MOMA/FBA solution vectors
    solutionVec = empty(len(model), 'd')
    # index of next job to be processed (as 0-dimensional numpy array)
    nxt = array(0, 'i')
    free_processes = set(range(1, numprocs))  # set of processes free for work
    nDone = 0  # number of finished jobs

    while nDone < nJobs:
        while free_processes and nxt < nJobs:
            # Assign reaction to free process
            print koGroups[nxt][0]
            proc_index = free_processes.pop()
            # Send index of next job to next free process
            comm.Send(nxt, proc_index)
            # Mark in job list (int() is used to convert from ref. to literal)
            jobList[proc_index] = int(nxt)
            nxt += 1

        # If there are more processes than jobs, stop extra processes
        while free_processes:
            comm.Send(stopMsg, free_processes.pop())

        # Receive results from compute processes
        comm.Recv(solutionVec, MPI.ANY_SOURCE, status=status)
        proc_index = status.Get_source()

        # Test if solutionVec[0] is nan (nan means not converged)
        if solutionVec[0] != solutionVec[0]:
            result[jobList[proc_index]] = (nan, nan, 0.)
        elif solutionVec[0] == inf:
            # inf means infeasible, i.e. no solution exists
            result[jobList[proc_index]] = (inf, inf, 0.)
        else:
            jobIndex = jobList[proc_index]
            solutionFlux = MetabolicFlux(model, solutionVec)
            diff = sum(solutionFlux.absDiff(wtSolution).fluxDict.values())
            obj_val = max(0., dot(objective, solutionVec))
            distance = moma.evalObjFunc(wtVec, solutionVec, weights)
            result[jobIndex] = (distance, diff, obj_val)
            if filePrefix:
                solutionFlux.writeToFile(filePrefix+koGroups[jobIndex][0]+
                                         _FILE_SUFFIX)

        # Increase job counter and mark reporting process as free
        nDone += 1
        free_processes.add(proc_index)

    # Stop last compute process
    while free_processes:
        comm.Send(stopMsg, free_processes.pop())

    return result


def parallel_ko_worker(comm, model, solver, wtVec, fbaParams, koGroups=None,
                       weights=None, numIter=1, useMoma=When.AUTO,
                       lethalityCutoff=_CUTOFF_LETHAL):
    """ compute process (worker) for parallel knockout analysis

    Keyword arguments:

    comm       -- MPI communicator
    model      -- the MetabolicModel
    solver     -- name of QP/NLP solver (or "default")
    wtVec      -- FBA solution for wildtype
    fbaParams  -- parameters for FBA (incl. linear equality and inequality
                                      constraints)
    koGroups   -- list of pairs (group name, list of reactions)
    weights    -- weight vector for weighted MOMA (None -> perform regular MOMA,
                  else: weight flux i with weights[i])
    numIter    -- number of NLP runs (if NLP solver is used) for MOMA
    useMoma    -- selector of optimization strategy, enum values (class When):
                  NEVER  - always perform FBA (LP)
                  AUTO   - perform FBA first, followed by MOMA if not lethal
                  ALWAYS - always perform MOMA (QP)
    lethalityCutoff
               -- threshold for biomass flux signifying lethality
    """
    if not koGroups:
        koGroups = _makeDefaultKoGroups(model)
#    print "worker: %d reactions, %d knockout groups" % (len(model),
#                                                        len(koGroups))
    wtSolution = MetabolicFlux(model, wtVec)
    nanVec = array([nan]*len(wtVec))
    infVec = array([inf]*len(wtVec))
    moma = MomaAnalyzer(solver)

    # index of reaction group to be knocked out (as 0-dimensional numpy array)
    i = array(0, 'i')
    comm.Recv(i)
    while i >= 0:
        group, reaList = koGroups[i]

        # Restrict flux through all reactions in group to zero
        tmp_lb = {}  # {index in model : LB value}
        tmp_ub = {}  # {  - '' -       : UB  ''  }
        for rea in reaList:
            try:
                rIndex = model.reactionDict[rea]
            except KeyError:
                # Skip any blocked reactions (knockout has no effect)
                continue
            tmp_lb[rIndex] = model.reactions[rIndex].lb
            tmp_ub[rIndex] = model.reactions[rIndex].ub
            model.reactions[rIndex].lb = model.reactions[rIndex].ub = 0.

        if not tmp_lb:
            print "  - skipped group '%s' (only blocked reactions)" % group
            comm.Send(wtVec)
            comm.Recv(i)
            continue

        obj_val = lethalityCutoff + 1.

        # Perform FBA first to check for lethality
        if useMoma != When.ALWAYS:
            fba = FbAnalyzer(fbaParams.solver)
            # Note: Model is already reduced
            obj_val, solutionFlux = fba.runOnModel(model, fbaParams,
                                                   rmDeadEnds=False)[:2]
            status = SolverStatus.PREPARED
            if not solutionFlux:
                obj_val = 0.

        # Perform MOMA if knockout not already predicted by FBA to be lethal
        if useMoma != When.NEVER and abs(obj_val) > lethalityCutoff:
            solutionFlux, status = moma.runOnModel(model, wtSolution,
                fbaParams.linConstraints, numIter, weights)[1:3]

        # Send result to dispatcher
        if status == SolverStatus.UNKNOWN:
            # If status is 'unknown, not converged', send vector with only 'nan'
            # entries
            comm.Send(nanVec)
        elif status in (SolverStatus.PRIM_INFEAS, SolverStatus.DUAL_INFEAS):
            # If status is 'infeasible', send vector with only 'inf' entries
            comm.Send(infVec)
        elif len(solutionFlux) == 0:
            if status == SolverStatus.PREPARED:
                # FBA infeasible
                comm.Send(infVec)
            else:
                comm.Send(nanVec)
        else:
            solutionVec = array(solutionFlux.getVecOrderedByModel(model))
            comm.Send(solutionVec)

        # Restore original constraints
        for rea in reaList:
            try:
                rIndex = model.reactionDict[rea]
            except KeyError:
                continue
            model.reactions[rIndex].lb = tmp_lb[rIndex]
            model.reactions[rIndex].ub = tmp_ub[rIndex]

        # Get next index
        comm.Recv(i)


def main():
    if _mpi_avail:
        comm = MPI.COMM_WORLD
        numprocs = comm.Get_size()
        rank = comm.Get_rank()
        if numprocs == 1:
            comm = None
    else:
        comm = None

    if not _mpi_avail or rank == 0:    # Start of root process code

        # 1. Parse command line

        usage = "Usage: %prog [options]"
        version = "Knockout analysis\n" + COPYRIGHT_VERSION_STRING
        parser = OptionParser(usage=usage, version=version)
        momaOptions = parser.add_option_group("General MOMA options")
        momaOptions.add_option("-r", "--reactions", dest="reactionFile",
                               help="perform knockout analysis on the network "
                               "given by the reaction FILE", metavar="FILE")
        momaOptions.add_option("-p", "--parameters", dest="paramFile",
                               help="use the given scenario FILE for the "
                               "wildtype", metavar="FILE")
        momaOptions.add_option("-o", "--output", dest="outputFile", help="write"
                               " output of knockout analysis to FILE",
                               metavar="FILE")
        momaOptions.add_option("-w", "--wt-solution", dest="wtSolution",
                               help="use the given solution FILE for the "
                               "wildtype (optional)", metavar="FILE")
        momaOptions.add_option("-v", "--reduce-by-fva", dest="redFvaFile",
                               help="use FVA result from FILE to reduce matrix "
                               "(experimental; only works if perturbed solution"
                               " space is a subset of the wildtype solution "
                               "space)", metavar="FILE")
        koOptions = parser.add_option_group("Options for knockout analysis")
        koOptions.add_option("-f", "--solution-files", dest="fluxFilePrefix",
                             help="write MOMA/FBA solutions to files marked "
                             "with FILE-PREFIX (may contain path)",
                             metavar="FILE-PREFIX")
        koOptions.add_option("-g", "--ko-groups", dest="koGroupDictFile",
                             help="read reaction groups to be knocked out "
                             "together from CSV FILE (if not given, individual "
                             "reactions are knocked out)", metavar="FILE")
        koOptions.add_option("-n", "--ko-ungrouped", action="store_true",
                             dest="koUngrouped", help="knock out individual "
                             "reactions that do not belong to any group (only "
                             "with -g)")
        koOptions.add_option("-m", "--moma", dest="useMoma", help="perform MOMA"
                             ", options: 'always', 'never' (FBA only), 'auto' "
                             "(default: FBA first, then MOMA if not lethal)",
                             metavar="WHEN")
        wmOptions = parser.add_option_group("Extra options for weighted MOMA")
        wmOptions.add_option("-x", "--wmoma", dest="wMomaFvaFile",
                             help="perform weighted MOMA with weights computed "
                             "from the given FVA solution FILE", metavar="FILE")
        wmOptions.add_option("-a", "--alpha", dest="alpha", type="float",
                             help="parameter ALPHA for computation of weights:"
                             " w = alpha + exp(-beta*(fvaMax-fvaMin)) (default:"
                              " %g)" % _DEFAULT_ALPHA)
        wmOptions.add_option("-b", "--beta", dest="beta", type="float",
                             help="parameter BETA for computation of weights: "
                             "w = alpha + exp(-beta*(fvaMax-fvaMin)) (default: "
                             "%g)" % _DEFAULT_BETA)
        optOptions = parser.add_option_group("Options for optimization")
        optOptions.add_option("-s", "--solver", dest="solver", help="QP solver "
                              "to be used for MOMA (default cvxopt)",
                              metavar="SOLVER")
        optOptions.add_option("-i", "--iterations", dest="numIter", type="int",
                              help="number of NLP runs to perform (from "
                              "different random start points) - per knockout")
        optOptions.add_option("-l", "--use-full-matrix", action="store_true",
                              dest="useFullMatrix", help="use full matrix "
                              "(disable removal of dead ends and nonfunctional "
                              "reactions)")
        parser.set_defaults(solver="default", numIter=1, koUngrouped=False,
                            useMoma="auto", useFullMatrix=False,
                            alpha=_DEFAULT_ALPHA, beta=_DEFAULT_BETA)

        try:
            options, _ = parser.parse_args()
            parser.check_required("-r")
            parser.check_required("-p")
            parser.check_required("-o")
        except SystemExit, e:
            _mpi_exit(comm, e.args[0])

        weighted = bool(options.wMomaFvaFile)
        if weighted:
            if options.alpha < 0.:
                print "Error: alpha must be non-negative."
                _mpi_exit(comm, 1)
            if options.beta <= 0.:
                print "Error: beta must be positive."
                _mpi_exit(comm, 1)

        if options.numIter < 1:
            print "Error: Number of NLP runs must be positive."
            _mpi_exit(comm, 1)

        try:
            useMoma = decodeWhen(options.useMoma, set((When.NEVER, When.AUTO,
                                                       When.ALWAYS)))
        except KeyError:
            print ("Error: Illegal value for -m option (allowed values are "
                   "'auto', 'never', or 'always').")
            _mpi_exit(comm, 1)

        # 2. Parse reaction file

        rparser = ReactionParser()
        model = MetabolicModel()
        try:
            model.addReactionsFromFile(options.reactionFile, rparser)
        except IOError, strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.reactionFile))
            print strerror
            _mpi_exit(comm, 1)
        except SyntaxError, strerror:
            print ("Error in reaction file %s:" %
                   os.path.basename(options.reactionFile))
            print strerror
            _mpi_exit(comm, 1)

        # 3. Parse scenario file

        model_messages = []
        pparser = ParamParser()
        try:
            maxmin, objStr, fbaSolver, numIter, lb , ub = \
                pparser.parse(options.paramFile)
            # Set flux bounds in model
            model.setFiniteBounds(lb, ub, True, model_messages)
            if fbaSolver == "":
                fbaSolver = "default"
            fbaParams = FbaParam(fbaSolver, maxmin, objStr, numIter)
            lin_constraints = pparser.lin_constraints
            fbaParams.setLinConstraints(lin_constraints)
        except IOError, strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.paramFile))
            print strerror
            _mpi_exit(comm, 1)
        except SyntaxError, strerror:
            print ("Error in scenario file %s:" %
                   os.path.basename(options.paramFile))
            print strerror
            _mpi_exit(comm, 1)
        except ValueError, strerror:
            print strerror
            _mpi_exit(comm, 1)

        # Show warning and info messages of parsers
        msgs = (rparser.getMessages() + pparser.getMessages() +
                [x[1] for x in model_messages])
        if msgs:
            print '\n'+'\n'.join(msgs)

        # 4. Read groups of reactions from file (if given)

        koGroups = []  # list of pairs (group name, list of reactions)

        if options.koGroupDictFile:
            koGroupDict = {}     # dict {group name : list of reactions}
            koGroupDictRev = {}  # dict {reaction : list of groups}
            koOrder = []         # order of reactions
            try:
                with open(options.koGroupDictFile) as f:
                    csvReader = csv.reader(f, delimiter=';')

                    for row in csvReader:
                        group, rea = map(str.strip, row[:2])
                        koOrder.append(rea)

                        if group in koGroupDict:
                            koGroupDict[group].append(rea)
                        else:
                            koGroupDict[group] = [rea]

                        if rea in koGroupDictRev:
                            koGroupDictRev[rea].append(group)
                        else:
                            koGroupDictRev[rea] = [group]

            except csv.Error as e:
                print ("Error in CSV file %s, line %d: %s" %
                       (os.path.basename(options.koGroupDictFile),
                        csvReader.line_num, e))
                _mpi_exit(comm, 1)
            except ValueError, strerror:
                print ("Error in CSV file %s, line %d: %s" %
                       (os.path.basename(options.koGroupDictFile),
                        csvReader.line_num, strerror))
                _mpi_exit(comm, 1)
            except IOError, strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(options.koGroupDictFile))
                print strerror
                _mpi_exit(comm, 1)

            # Generate list of knockouts (order as close as possible to order of
            #                             reactions in file)
            processedGroups = set()
            for rea in koOrder:
                for group in koGroupDictRev[rea]:

                    if group in ("", "NULL"):
                        if options.koUngrouped:
                            # Create new group "KO_x" for individual reaction x
                            koGroups.append(("KO_"+rea, [rea]))
                        # else skip

                    elif group not in processedGroups:
                        koGroups.append((group, koGroupDict[group]))
                        processedGroups.add(group)

        else:
            # Knock out individual reactions
            koGroups = _makeDefaultKoGroups(model)

        # 5. (Optionally) Read FVA solution for reducing the solution space
        #    - skip if full matrix is to be used

        if options.redFvaFile and not options.useFullMatrix:
            try:
                fbaSolution, redMinmax = \
                    FvAnalyzer.parseSolutionFile(options.redFvaFile)
            except IOError, strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(options.redFvaFile))
                print strerror
                _mpi_exit(comm, 1)
            except SyntaxError, strerror:
                print ("Error in FVA solution file %s:" %
                       os.path.basename(options.redFvaFile))
                print strerror
                _mpi_exit(comm, 1)

            if not fbaSolution.hasSameReactions(model):
                print ("Error in file %s: FVA solution and model must have the "
                       "same reactions." % os.path.basename(options.redFvaFile))
                _mpi_exit(comm, 1)

            # Determine blocked reactions based on FVA solution
            blockedReactions = FvAnalyzer.getBlockedReactions(redMinmax)

        elif options.useFullMatrix:
            blockedReactions = []
        else:
            # Determine blocked reactions based on dead-end analysis
            blockedReactions = model.findDeadEnds(True)[1]

        # 6. (Optionally) Read FVA solution and compute weights for wMOMA

        if weighted:
            try:
                fbaSolution, wmomaMinmax = \
                    FvAnalyzer.parseSolutionFile(options.wMomaFvaFile)
            except IOError, strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(options.wMomaFvaFile))
                print strerror
                _mpi_exit(comm, 1)
            except SyntaxError, strerror:
                print ("Error in FVA solution file %s:" %
                       os.path.basename(options.wMomaFvaFile))
                print strerror
                _mpi_exit(comm, 1)

            if not fbaSolution.hasSameReactions(model):
                print ("Error in file %s: FVA solution and model must have the "
                     "same reactions." % os.path.basename(options.wMomaFvaFile))
                _mpi_exit(comm, 1)

            weights = MomaAnalyzer.getWeightsFromFluxVar(model, wmomaMinmax,
                                                    options.alpha, options.beta)

        # 7.a Read wildtype solution from file (if given)

        if options.wtSolution:
            # Read wildtype solution from file
            wtSolution = MetabolicFlux()
            try:
                wtSolution.readFromFile(options.wtSolution)
            except IOError, strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(options.wtSolution))
                print strerror
                _mpi_exit(comm, 1)
            except SyntaxError, strerror:
                print ("An error occurred parsing file %s:" %
                       os.path.basename(options.wtSolution))
                print strerror
                _mpi_exit(comm, 1)

            if not wtSolution.hasSameReactions(model):
                print ("Error in file %s: FBA solution and model must have the "
                       "same reactions." % os.path.basename(options.wtSolution))
                _mpi_exit(comm, 1)

            # Build coefficient vector of linear objective function
            objective = array(ParamParser.convertObjFuncToLinVec(objStr,
                                                            model.reactionDict))
            wt_obj_value = dot(objective,
                               wtSolution.getVecOrderedByModel(model))

        # 7.b If FBA solution is not given, perform FBA

        else:
            # Compute wildtype solution via FBA
            fba = FbAnalyzer(fbaSolver)

            wt_obj_value, wtSolution, ndim = fba.runOnModel(model, fbaParams,
                rmDeadEnds=not options.useFullMatrix)
            if not options.useFullMatrix:
                print ("Info: The reduced network for FBA has %u reactions and "
                       "%u metabolites." % ndim[1::-1])
            if len(wtSolution) == 0:
                print "Model is infeasible or unbounded. Nothing to do."
                _mpi_exit(comm)

        solver = options.solver
        if not solver:
            solver = "default"
        numIter = options.numIter

        # 8. Generate submodel with blocked reactions removed

        if blockedReactions:
            modelRed = model.getSubModelByExcludeList(blockedReactions)
            reactionsRed = set(modelRed.getReactionNames())
            # Reduce solution to non-dead reactions
            wtSolutionRed = MetabolicFlux()
            for rea in wtSolution:
                if rea in reactionsRed:
                    wtSolutionRed.fluxDict[rea] = wtSolution.fluxDict[rea]
                    wtSolutionRed.boundsDict[rea] = wtSolution.boundsDict[rea]
            # Also reduce weights vector
            if weighted:
                weightsRed = [0.]*len(modelRed)
                for rea in wtSolution:
                    if rea in reactionsRed:
                        weightsRed[modelRed.reactionDict[rea]] = \
                            weights[model.reactionDict[rea]]
                weightVec = array(weightsRed)
            else:
                weightVec = None
        else:
            modelRed = model
            wtSolutionRed = wtSolution
            if weighted:
                weightVec = array(weights)
            else:
                weightVec = None

        modelStr = modelRed.writeToString(False, False, False)
        nReactions = len(modelRed)
        lb, ub = map(array, modelRed.getBounds())
        # Build coefficient vector of linear objective function
        objective = array(ParamParser.convertObjFuncToLinVec(objStr,
                                                        modelRed.reactionDict))
        if _mpi_avail and numprocs > 1:
            intBuf = array((nReactions, numIter, int(useMoma), int(maxmin),
                            int(weighted)), 'i')
            comm.Bcast(intBuf)
            wtVec = array(wtSolutionRed.getVecOrderedByModel(modelRed))
            if weighted:
                weightVec = array(weightsRed)
            else:
                weightVec = None

    else:                              # Start of compute process (slave) code
        solver = None
        modelStr = None
        intBuf = empty(5, 'i')
        comm.Bcast(intBuf)
        nReactions, numIter, useMoma, maxmin, weighted = intBuf
        lb = empty(nReactions, 'd')
        ub = empty(nReactions, 'd')
        wtVec = empty(nReactions, 'd')
        fbaSolver = None
        objStr = None
        lin_constraints = None
        koGroups = None
        if weighted:
            weightVec = empty(nReactions, 'd')
        else:
            weightVec = None

    # 9. Sequentially knock out fluxes and perform MOMA against the wildtype
    #    solution

    if _mpi_avail and numprocs > 1:
        # Launch parallel analysis
        solver = comm.bcast(solver)
        modelStr = comm.bcast(modelStr)
        comm.Bcast(lb)
        comm.Bcast(ub)
        comm.Bcast(wtVec)
        fbaSolver = comm.bcast(fbaSolver)
        objStr = comm.bcast(objStr)
        fbaParams = FbaParam(fbaSolver, maxmin, objStr)
        fbaParams.linConstraints = comm.bcast(lin_constraints)
        koGroups = comm.bcast(koGroups)
        if weighted:
            comm.Bcast(weightVec)

        if rank == 0:
            result = parallel_ko_dispatch(comm, numprocs, modelRed, solver,
                            objective, wtSolutionRed, wt_obj_value, koGroups,
                            options.fluxFilePrefix, weightVec)
        else:
            model = MetabolicModel()
            model.addReactionsFromString(modelStr)
            model.setBounds(lb, ub)
            parallel_ko_worker(comm, model, solver, wtVec, fbaParams, koGroups,
                               weightVec, numIter, useMoma)

    else:
        # Launch serial analysis
        result = serial_knockout(modelRed, solver, objective, wtSolutionRed,
                                 fbaParams, koGroups, options.fluxFilePrefix,
                                 wt_obj_value, weightVec, numIter, useMoma)

    if not _mpi_avail or rank == 0:

    # 10. Write output to file

        print "Writing output file."
        try:
            write_output(options.outputFile, [x[0] for x in koGroups], result,
                         wt_obj_value)
        except IOError, strerror:
            print ("Unable to write to file %s:" %
                   os.path.basename(options.outputFile))
            print strerror
            _mpi_exit(comm, 1)


def write_output(filename, group_names, result, wt_obj_val):
    # Construct output table and get widths of table columns (length of
    # longest occurring string)
    output = []
    for i in range(len(group_names)):
        output.append((group_names[i],) + tuple(map(padNumber, map(repr,
            (100.*result[i][2]/wt_obj_val,) + result[i]))))
        maxlen = tuple(map(len, output[-1]))

    # Create list of (index, value of objective function) pairs, sorted by value
    # in descending order - for ordering the rows of the output table
    order = sorted(zip(range(len(result)), [x[2] for x in result]),
                   key=lambda x : x[1], reverse=True)

    with open(filename, 'w') as f:
        # Write table head
        f.write("NAME".ljust(maxlen[0])+"   "+"OBJ_FUN(%)".center(maxlen[1])+" "
                +"DISTANCE".rjust(maxlen[2])+" "+"ABS_DIFF".rjust(maxlen[3])+" "
                +"OBJ_FUN_VAL".rjust(maxlen[4])+"\n")

        # Write table rows
        for i, _ in order:
            reaName, objfPercent, dist, diff, objfValue = output[i]
            f.write(reaName.ljust(maxlen[0]) + " : " +
                    objfPercent.rjust(maxlen[1]) + " " + dist.rjust(maxlen[2]) +
                    " " + diff.rjust(maxlen[3]) + " " +
                    objfValue.rjust(maxlen[4]) + "\n")


if __name__ == "__main__":
    main()



"""
Convex optimization modeling for cvxopt.

"""

# Copyright (C) 2006-2008 Jacob Mattingley and Stephen Boyd.
#
# This file is part of CVXMOD.
#
# CVXMOD is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# CVXMOD is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["OptimizationError", "DimError", "OptvarValueError",
           "ParamValueError", "ProductError", "AssertionError",
           "ClassificationError", "ConvexityError",
           "ConstraintComplexityError", "StdFormError", "AtomArgsError",
           "TestError", "TransposedVarError", "OptvarBoundError"]

class OptimizationError(Exception):
    """Base class for errors in the jo package."""
    pass

class DimError(OptimizationError):
    """Occurs when combining two expressions of incompatible dimension."""
    def __init__(self, value='expressions have incompatible dimensions'):
        self.value = value

    def __str__(self):
        return repr(self.value)

class OptvarValueError(OptimizationError):
    """Occurs when an optvar has value None when it should not."""
    def __init__(self, value="optvar has value None"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ParamValueError(OptimizationError):
    """Occurs when a param has an inappropriate value (e.g. None when it
    shouldn't be."""
    def __init__(self, value="invalid param value"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ProductError(OptimizationError):
    """Occurs when two variables are multiplied."""
    def __init__(self, value="cannot form product of optimization variables"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class AssertionError(OptimizationError):
    """Occurs when a param assertion is not met."""
    def __init__(self, value="problem assertion failed"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ClassificationError(OptimizationError):
    """Occurs when convexity or monotonicity cannot be determined."""
    def __init__(self, value="cannot classify object"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ConvexityError(OptimizationError):
    """Occurs when convexity is required and not assured."""
    def __init__(self, value="could not determine required convexity"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ConstraintComplexityError(OptimizationError):
    """Occurs when a constraint cannot be transformed or understood."""
    def __init__(self, value="constraint is too complicated"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class StdFormError(OptimizationError):
    """Occurs when an expression is not in the expected standard form."""
    def __init__(self, value="expression is not in standard form"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class AtomArgsError(OptimizationError):
    """Occurs when an atom is passed incorrect arguments."""
    def __init__(self, value="atom was passed incorrect arguments"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class TestError(OptimizationError):
    """Occurs when an error is encountered during testing."""
    def __init__(self, value="error occurred during testing"):
        self.value = value

    def __str__(self):
        return repr(self.value)

class TransposedVarError(OptimizationError):
    """Occurs when a variable is transposed in an expression, thus preventing a
    correct result."""
    def __init__(self, value="transposed variable prevents valid result"):
        self.value = value

    def __str__(self):
        return repr(self.value)
class OptvarBoundError(OptimizationError):
    """Occurs when an optvar does not have a required bound."""
    def __init__(self, value="optvar does not have bound"):
        self.value = value

    def __str__(self):
        return repr(self.value)



#!/usr/bin/env python

""" This script checks whether the given flux distribution (FBA solution file)
    conforms to the constraints imposed by the given reaction and/or scenario
    file. It lists all constraints violations. There is zero tolerance, e.g. a
    value of -1E-25 violates an "LB 0." constraint.

Usage:

    to_check_constraints.py <solution-file> [-r reaction-file] [-p param-file]


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

from fba import OptionParser
from reactionparser import ReactionParser
from paramparser import ParamParser
from metabolicmodel import MetabolicModel
from metabolicflux import MetabolicFlux
from defines import COPYRIGHT_VERSION_STRING
import os
import sys
from numpy import dot


def checkIrrev(solution, model):
    """ check the signs of the fluxes in 'solution' associated with irreversible
        reactions in the given model

    Returns negViolations, posViolations - {reaction : flux} dicts
    """
    negViolations, posViolations =  {}, {}
    for rea in solution:
        if rea not in model.reactionDict:
            continue
        rea_index = model.reactionDict[rea]
        flux = solution[rea]

        if model.reactions[rea_index].lb >= 0. and flux < 0.:
            negViolations[rea] = flux
        elif model.reactions[rea_index].ub <= 0. and flux > 0.:
            posViolations[rea] = flux

    return negViolations, posViolations


def checkBoundsExplicit(solution, lb, ub):
    """ check if flux distribution 'solution' violates any of the given lower
        and upper flux bounds

    Keyword arguments:

    solution -- flux distribution (MetabolicFlux object)
    lb, ub   -- dicts {reaction : bound} of lower and upper bounds

    Returns:  lbViolations, ubViolations - {reaction : flux, bound} dicts
    """
    lbViolations, ubViolations = {}, {}
    for rea in solution:
        flux = solution[rea]

        if rea in lb:
            if flux < lb[rea]:
                lbViolations[rea] = flux, lb[rea]
        if rea in ub:
            if flux > ub[rea]:
                ubViolations[rea] = flux, ub[rea]

    return lbViolations, ubViolations


def checkBoundsByModel(solution, model):
    """ check if the flux distribution 'solution' violates any flux bounds
        defined for the reactions in the given model

    Returns:  lbViolations, ubViolations - {reaction : flux, bound} dicts
    """
    lbViolations, ubViolations = {}, {}
    for rea in solution:
        if rea not in model.reactionDict:
            continue
        rea_index = model.reactionDict[rea]
        flux = solution[rea]
        lb, ub = model.reactions[rea_index].lb, model.reactions[rea_index].ub

        if flux < lb:
            lbViolations[rea] = flux, lb
        elif flux > ub:
            ubViolations[rea] = flux, ub

    return lbViolations, ubViolations


def checkBoundsInSolution(solution):
    """ check if flux distribution 'solution' violates any of the flux bounds
        associated with its reactions

    Returns:  lbViolations, ubViolations - {reaction : flux, bound} dicts
    """
    lbViolations, ubViolations = {}, {}
    for rea in solution:
        flux = solution[rea]
        if rea not in solution.boundsDict:
            continue
        lb, ub = solution.boundsDict[rea]

        if flux < lb:
            lbViolations[rea] = flux, lb
        elif flux > ub:
            ubViolations[rea] = flux, ub

    return lbViolations, ubViolations


def checkMassBalance(solution, model):
    """ check if flux distribution 'solution' violates the mass balance
        constraints defined by the model's stoichiometric matrix

    Returns:
    avgErr, maxPosErr, minNegErr, maxRowMet, minRowMet

    avgErr    -- average deviation from zero over all rows
    maxPosErr -- highest positive deviation from zero
    minNegErr -- lowest negative deviation from zero
    maxRowMet -- row (metabolite name) where maxPosErr occurred
    minRowMet -- row (metabolite name) where minNegErr occurred
    """
    solutionVec = solution.getVecOrderedByModel(model)
    sv = dot(model.getStoichiometricMatrix(), solutionVec)
    svmax = svmin = 0.
    svmax_i = svmin_i = -1
    svsum = 0.
    nMetabolites = len(sv)
    for i in range(nMetabolites):
        svsum += abs(sv[i])
        if sv[i] < svmin:
            svmin = sv[i]
            svmin_i = i
        elif sv[i] > svmax:
            svmax = sv[i]
            svmax_i = i

    return (svsum / nMetabolites, svmax, svmin, model.metabolites[svmax_i],
                                                model.metabolites[svmin_i])

def printMassBalanceCheckResult(svavg, svmax, svmin, svmaxMet, svminMet):
    """ print the result of checkMassBalance() - simple call:
        printMassBalanceCheckResult(*checkMassBalance(solution, model))
    """
    print "Mean deviation from zero in S*v: %g" % svavg
    print ("Maximum pos. deviation from zero in S*v: %g in metabolite %s" %
           (svmax, svmaxMet))
    print ("Maximum neg. deviation from zero in S*v: %g in metabolite %s" %
           (svmin, svminMet))


def checkLinConstraints(solution, model, linConstraints):
    """ check if flux distribution 'solution' violates any of the given linear
        equality and inequality constraints

    Keyword arguments:

    solution       -- flux distribution (MetabolicFlux object)
    model          -- MetabolicModel (for generating flux and coefficient
                                      vectors)
    linConstraints -- list of LinearConstraint objects

    Returns:
    eqErr, ineqErr

    eqErr   -- dict {index in linConstraints : deviation} for each equality
               constraint
    ineqErr -- dict {index in linConstraints : deviation} for each inequality
               constraint - if deviation is negative, constraint is satisfied
    """
    solutionVec = solution.getVecOrderedByModel(model)
    eqs, ineqs = ParamParser.linConstraintsToVectors(linConstraints,
                                                     model.reactionDict)
    eqIndex, ineqIndex = [], []
    for i in range(len(linConstraints)):
        if linConstraints[i].isEq:
            eqIndex.append(i)
        else:
            ineqIndex.append(i)

    eqErr, ineqErr = {}, {}
    if eqs:
        A = [row[0] for row in eqs]
        b = [row[1] for row in eqs]
        dotVec = dot(A, solutionVec)
        for i in range(len(b)):
            eqErr[eqIndex[i]] = abs(dotVec[i]-b[i])

    if ineqs:
        A = [row[0] for row in ineqs]
        b = [row[1] for row in ineqs]
        dotVec = dot(A, solutionVec)
        for i in range(len(b)):
            ineqErr[ineqIndex[i]] = dotVec[i]-b[i]

    return eqErr, ineqErr

def printLinConstraintCheckResult(eqErr, ineqErr, linConstraints):
    if eqErr:
        print "Deviation in additional linear equality constraints:"
        j = 1  # Counter for enumeration of equality constraints
        for i in sorted(eqErr):
            print "%3d: %g (%s)" % (j, eqErr[i], linConstraints[i])
            j += 1

    if ineqErr:
        conforms = True
        j = 1  # Counter for enumeration of violated inequality constraints
        for i in sorted(ineqErr):
            if ineqErr[i] > 0.:
                if conforms:
                    print "Deviation in linear inequality constraints:"
                    conforms = False
                print "%3d: %g (%s)" % (j, ineqErr[i], linConstraints[i])
        if conforms:
            print ("The given solution conforms to all linear inequality "
                   "constraints.")


def main():
    # 1. Parse command line

    usage = "Usage: %prog <solution-file> [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile", help="use "
                      "reversibilities from reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="use constraints from the given scenario FILE",
                      metavar="FILE")

    options, args = parser.parse_args()

    # 2. Read solution file

    solution = MetabolicFlux()
    try:
        solution.readFromFile(args[0])
    except IndexError:
        print ("Error: No solution file given.\nUsage is\n    " +
               os.path.basename(sys.argv[0]) + " <solution-file> [options]")
        exit()
    except IOError, strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(args[0]))
        print strerror
        exit()
    except SyntaxError, strerror:
        print ("An error occurred parsing file %s:" %
               os.path.basename(args[0]))
        print strerror
        exit()

    if options.reactionFile is None and options.paramFile is None:
        print "Neither reaction nor scenario file given. Nothing to do."
        exit()

    # 3. Parse reaction file (if given) and check irreversibilities

    if options.reactionFile:
        model = MetabolicModel()
        rparser = ReactionParser()
        try:
            model.addReactionsFromFile(options.reactionFile, rparser)
        except IOError, strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.reactionFile))
            print strerror
            exit()
        except SyntaxError, strerror:
            print ("Error in reaction file %s:" %
                   os.path.basename(options.reactionFile))
            print strerror
            exit()

        negViolations, posViolations = checkIrrev(solution, model)
        conforms = not negViolations and not posViolations

        for rea in negViolations:
            print ("Irreversible reaction %s has negative flux (%g)." %
                   (rea, negViolations[rea]))
        for rea in posViolations:
            print ("Flipped irreversible reaction %s has positive flux (%g)." %
                   (rea, posViolations[rea]))
    else:
        conforms = True
        model = None

    # 4. Parse scenario file (if given) and check LB/UB constraints

    if options.paramFile:
        pparser = ParamParser()
        try:
            lb, ub = pparser.parse(options.paramFile)[4:]
        except IOError, strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(options.paramFile))
            print strerror
            exit()
        except SyntaxError, strerror:
            print ("Error in scenario file %s:" %
                   os.path.basename(options.paramFile))
            print strerror
            exit()
        except ValueError, strerror:
            print strerror
            exit()
        linConstraints = pparser.lin_constraints

        lbViolations, ubViolations = checkBoundsExplicit(solution, lb, ub)
        conforms = conforms and not lbViolations and not ubViolations

        for rea in lbViolations:
            print ("%s (%g) violates LB %g." % ((rea,)+ lbViolations[rea]))
        for rea in ubViolations:
            print ("%s (%g) violates UB %g." % ((rea,)+ ubViolations[rea]))
    else:
        linConstraints = None

    if conforms:
        print ("The given solution does not violate any bounds.")

    if model:

        # 5. Check mass balance constraints

        printMassBalanceCheckResult(*checkMassBalance(solution, model))

        # 6. Check additional equality and inequality constraints

        printLinConstraintCheckResult(*checkLinConstraints(solution, model,
                                                           linConstraints),
                                      linConstraints=linConstraints)


if __name__ == "__main__":
    main()



"""
This module defines the class TransporterParser, a parser for transporter files.


This file is part of metano.
Copyright (C) 2010-2014 Alexander Riemer,
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

import re

class TransporterParser(object):
    """Parser for transporter files.
    It reads lines of the format

    <name> [<factor>] [co(<name>)]

    into a list. Lines beginning with a # sign are ignored
    """

    # line pattern : <name> [<factor>] [co(<name>)]
    line_pattern = re.compile(r"(\S*)(?:\s*([\d\.]*))?(?:\s*co\((\S*)\))?")

    # name pattern: *_<type>_* (e.g. "trans_c_glu" => type "c")
    name_pattern = re.compile(r"[^_]*_([^_]*)_")


    def __init__(self):
        """ initialize the TransporterParser
        """
        self.clear()


    def clear(self):
        """ clear all internal variables, warnings and info messages
        """
        self.transporters = []


    def parse(self, filename):
        """ parse transporter file

        Keyword arguments:

        filename -- name of the file to be parsed

        Returns: list of (transporter_name, flux_factor, cotransporter_name)

        cotransporter_name is None if no cotransporter is to be checked
        flux_factor is None if the transporter is not of the "main" type (e.g.
        if it is not a C source)
        """
        try:
            line_no = 0
            with open(filename) as f:
                for line in f:
                    line_no += 1
                    line = line.strip()
                    if line == "" or line[0] == "#":
                        continue  # skip blank and comment lines
                    m = self.line_pattern.match(line)
                    transname = m.group(1)
                    transfac = m.group(2)
                    if transfac == "":
                        transfac = None
                    else:
                        transfac = float(transfac)
                    cotrans = m.group(3)
                    self.transporters.append((transname, transfac, cotrans))

        except ValueError:
            raise ValueError("Illegal float value %s in line %u of transporter "
                             "file." % (transfac, line_no))
        except AttributeError:
            raise SyntaxError("Syntax error in line %u of transporter file." %
                              line_no)

        return self.transporters


    def split_by_type(self, l=None):
        """ split transporters by type (e.g. element source - C, N, etc.)

        Keyword arguments:

        l -- list of (transporter_name, flux_factor, cotransporter_name),
                self.transporters (generated by parse()) is used if l is None

        Returns: list of lists of (type, transporter_name, flux_factor,
                 cotransporter_name), where type is the same for every
                 individual list
        """
        if l is None:
            l = self.transporters

        dict_of_types = {}  # dictionary { type : index in list_of_lists }
        list_of_lists = []
        try:
            for entry in l:
                m = self.name_pattern.match(entry[0])
                typ = m.group(1)
                if typ not in dict_of_types:
                    dict_of_types[typ] = len(list_of_lists)
                    list_of_lists.append([(typ,)+entry])
                else:
                    list_of_lists[dict_of_types[typ]].append((typ,)+entry)

        except AttributeError:
            raise SyntaxError("Malformed transporter name: %s (cannot identify "
                              "typ)" % entry[0])

        return list_of_lists
